#!/usr/bin/env python
"""Count the number of :term:`read alignments<alignment>` covering 
regions of interest in the genome, and calculate read densities (in reads
per nucleotide and in :term:`RPKM`) over these regions.

Results are output as a table with the following columns:

    ========================  ==================================================
    **Name**                  **Definition**
    ------------------------  --------------------------------------------------
    `region_name`             Name or ID of region of interest
    
    `region`                  Genomic coordinates of region, formatted as
                              described in
                              :meth:`plastid.genomics.roitools.SegmentChain.from_str`
                              
    `counts`                  Number of reads mapping to region
    
    `counts_per_nucleotide`   Read density, measured in number of reads mapping
                              to region, divided by length of region
                              
    `rpkm`                    Read density, measured in :term:`RPKM`
    
    `length`                  Region length, in nucleotides
    ========================  ==================================================
    
If a :term:`mask annotation file` is supplied, masked portions of regions
will be excluded when tabulating counts, measuring region length, and calculating
`counts_per_nucleotide` and `rpkm`.


See also
--------
:mod:`~plastid.bin.cs` script
    Count the number of :term:`read alignments<alignment>` and calculate
    read densities (in :term:`RPKM`) specifically for genes and sub-regions
    (5\' UTR, CDS, 3\' UTR), excluding positions covered by multiple genes
"""
import argparse
import inspect
import sys
import itertools
import warnings
import numpy

from plastid.genomics.roitools import SegmentChain
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import argsopener, get_short_name
from plastid.util.scriptlib.argparsers import (
    AnnotationParser,
    AlignmentParser,
    MaskParser,
    BaseParser,
)
from plastid.util.scriptlib.help_formatters import format_module_docstring

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

_DISABLED = ["normalize"]


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    ap = AnnotationParser()
    annotation_file_parser = ap.get_parser(conflict_handler="resolve")

    al = AlignmentParser(disabled=_DISABLED)
    alignment_file_parser = al.get_parser(conflict_handler="resolve")

    mp = MaskParser()
    mask_file_parser = mp.get_parser()

    bp = BaseParser()
    base_parser = bp.get_parser()

    parser = argparse.ArgumentParser(
        description     = format_module_docstring(__doc__),
        formatter_class = argparse.RawDescriptionHelpFormatter,
        parents = [
            base_parser,
            alignment_file_parser,
            annotation_file_parser,
            mask_file_parser,
        ],
    ) # yapf: disable
    parser.add_argument("outfile", type=str, help="Output filename")

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    ga = al.get_genome_array_from_args(args, printer=printer)
    transcripts = ap.get_transcripts_from_args(args, printer=printer, return_type=SegmentChain)
    crossmap = mp.get_genome_hash_from_args(args, printer=printer)

    ga_sum = ga.sum()
    normconst = 1000.0 * 1e6 / ga_sum

    with argsopener(args.outfile, args, "w") as fout:
        fout.write("## total_dataset_counts: %s\n" % ga_sum)
        fout.write("region_name\tregion\tcounts\tcounts_per_nucleotide\trpkm\tlength\n")
        for n, ivc in enumerate(transcripts):
            name = ivc.get_name()
            masks = crossmap.get_overlapping_features(ivc)
            ivc.add_masks(*itertools.chain.from_iterable((X for X in masks)))
            if n % 1000 == 0:
                printer.write("Processed %s regions..." % n)

            counts = numpy.nansum(ivc.get_masked_counts(ga))
            length = ivc.masked_length
            rpnt = numpy.nan if length == 0 else float(counts) / length
            rpkm = numpy.nan if length == 0 else rpnt * normconst
            ltmp = [name, str(ivc), "%.8e" % counts, "%.8e" % rpnt, "%.8e" % rpkm, "%d" % length]
            fout.write("%s\n" % "\t".join(ltmp))

        fout.close()

    printer.write("Processed %s regions total." % n)

    printer.write("Done.")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Exports a table of parent-child feature relationships for all feature types
found in a `GFF3`_ file. Features with multiple parents are sorted into a category
called `Multiple`.
"""
from plastid.readers.gff import GFF3_Reader
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import get_short_name, opener, argsopener
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.scriptlib.argparsers import BaseParser
from collections import Counter
import argparse
import sys
import inspect
import warnings

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def main(argv=sys.argv[1:]):
    """Command-line program
	
	Parameters
	----------
	argv : list, optional
		A list of command-line arguments, which will be processed
		as if the script were called from the command line if
		:py:func:`main` is called directly.

		Default: `sys.argv[1:]`. The command-line arguments, if the script is
		invoked from the command line
	"""
    bp = BaseParser()
    parser = argparse.ArgumentParser(
        description     = format_module_docstring(__doc__),
        formatter_class = argparse.RawDescriptionHelpFormatter,
        parents         = [bp.get_parser()]
    )
    parser.add_argument(
        "--exclude", nargs="+", default=[], help="Feature types to exclude from consideration"
    )
    parser.add_argument("infile",  metavar="infile.gff", type=str, help="Input GFF3 file")
    parser.add_argument("outfile", metavar="outfile.txt", type=str, help="Name of output file")

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)
    excluded = set(args.exclude)

    fin = sys.stdin if args.infile == "-" else opener(args.infile)

    # yapf : disable
    feature_counts        = Counter()
    features_with_parents = []
    feature_types         = {}
    name_type             = {}
    # yapf : enable

    printer.write("Opening %s..." % args.infile)
    c = 0
    for feature in GFF3_Reader(fin, return_stopfeatures=False):
        if c % 10000 == 0:
            printer.write("Processed %s features..." % c)
        c += 1
        ftype = feature.attr["type"]
        fname = feature.get_name()
        if ftype not in excluded:
            if ftype not in feature_types:
                feature_types[ftype] = Counter()
            feature_counts[ftype] += 1
            if fname is not None:
                name_type[fname] = ftype
            if "Parent" in feature.attr:
                features_with_parents.append(feature)
            else:
                feature_types[ftype]["parent unspecified"] += 1

    printer.write("Sorting parents...")
    c = 0
    for feature in features_with_parents:
        if c % 10000 == 0:
            printer.write("Processed %s parents..." % c)
        c += 1
        pnames = feature.attr["Parent"]
        ftype = feature.attr["type"]
        if pnames == "":
            feature_types[ftype]["parent unspecified"] += 1
        else:
            if len(pnames) > 1:
                feature_types[ftype]["multiple parents"] += 1
            else:
                ptype = name_type.get(pnames[0], "parent not in database")
                feature_types[ftype][ptype] += 1

    rows = sorted(feature_types.keys())
    cols = rows + ["parent unspecified", "parent not in database", "multiple parents"]

    with argsopener(args.outfile, args, "w") as fh:
        printer.write("Writing %s..." % args.outfile)
        header = "#feature_type\tcount\t" + "\t".join(cols) + "\n"
        fh.write(header)
        for r in rows:
            sout = "%s\t%s" % (r, feature_counts[r])
            for i in cols:
                sout += "\t%s" % feature_types[r].get(i, 0)
            fh.write("%s\n" % sout)

    printer.write("Done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Tests for methods in :py:mod:`plastid.bin.crossmap`
"""
import unittest
from plastid.util.services.mini2to3 import cStringIO
from nose.plugins.attrib import attr
from Bio import SeqIO
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from plastid.util.services.exceptions import MalformedFileError
from plastid.bin.crossmap import simulate_reads, \
                                     FastaNameReader, \
                                     revcomp_mask_chain, \
                                     fa_to_bed

#===============================================================================
# INDEX: test case
#===============================================================================


@attr(test="unit")
class TestCrossmap(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ivcs = {
            "plus": [
                SegmentChain(GenomicSegment("chrA", 0, 100, "+")),
                SegmentChain(GenomicSegment("chrA", 50, 100, "+")),
                SegmentChain(GenomicSegment("chrA", 50, 51, "+"))
            ],
            "minus_k25_off0": [
                SegmentChain(GenomicSegment("chrA", 0 + 25 - 1, 100 + 25 - 1, "-")),
                SegmentChain(GenomicSegment("chrA", 50 + 25 - 1, 100 + 25 - 1, "-")),
                SegmentChain(GenomicSegment("chrA", 50 + 25 - 1, 51 + 25 - 1, "-"))
            ],
            "minus_k50_off0": [
                SegmentChain(GenomicSegment("chrA", 0 + 50 - 1, 100 + 50 - 1, "-")),
                SegmentChain(GenomicSegment("chrA", 50 + 50 - 1, 100 + 50 - 1, "-")),
                SegmentChain(GenomicSegment("chrA", 50 + 50 - 1, 51 + 50 - 1, "-"))
            ],
            "minus_k25_off10": [
                SegmentChain(
                    GenomicSegment("chrA", 0 + 25 - 1 - 2 * 10, 100 + 25 - 1 - 2 * 10, "-")
                ),
                SegmentChain(
                    GenomicSegment("chrA", 50 + 25 - 1 - 2 * 10, 100 + 25 - 1 - 2 * 10, "-")
                ),
                SegmentChain(
                    GenomicSegment("chrA", 50 + 25 - 1 - 2 * 10, 51 + 25 - 1 - 2 * 10, "-")
                )
            ],
            "minus_k50_off10": [
                SegmentChain(
                    GenomicSegment("chrA", 0 + 50 - 1 - 2 * 10, 100 + 50 - 1 - 2 * 10, "-")
                ),
                SegmentChain(
                    GenomicSegment("chrA", 50 + 50 - 1 - 2 * 10, 100 + 50 - 1 - 2 * 10, "-")
                ),
                SegmentChain(
                    GenomicSegment("chrA", 50 + 50 - 1 - 2 * 10, 51 + 50 - 1 - 2 * 10, "-")
                )
            ],
        }

    def test_simulate_reads(self):
        fh = cStringIO.StringIO()
        genome = SeqIO.parse(cStringIO.StringIO(SHORT_FASTA), "fasta")
        for seq in genome:
            simulate_reads(seq, k=25, fh=fh)
        fh.seek(0)
        reads = fh.read()
        self.assertEqual(reads, SHORT_FASTA_KMERS)

    def test_fasta_name_reader(self):
        # make sure we get out only names of sequences from FASTA file
        reader = FastaNameReader(cStringIO.StringIO(SHORT_FASTA))
        self.assertEqual(next(reader), "chr50a")
        self.assertEqual(next(reader), "chr30b")
        self.assertRaises(StopIteration, reader.next)

    def test_revcomp_mask_chain_no_offset(self):
        for k in (25, 50):
            for offset in (0, 10):
                revlist = [revcomp_mask_chain(X, k, offset) for X in self.ivcs["plus"]]
                self.assertListEqual(self.ivcs["minus_k%s_off%s" % (k, offset)], revlist)

    def test_fa_to_bed(self):
        # test with and without offsets
        # start with block of FASTA formatted sequence:
        #     two blocks on same chromosome
        #     additional block on next chromosome
        #     additional block on next chromosome
        # two blocks on chrA, one on chrB
        # test without offset
        self.maxDiff = None

        reader = cStringIO.StringIO(CROSSMAP_BLOCK)

        # test without offset
        blocks1 = list(fa_to_bed(reader, 25, offset=0))
        self.assertListEqual(blocks1, CROSSMAP1)

        # test with offset
        reader = cStringIO.StringIO(CROSSMAP_BLOCK)
        blocks2 = list(fa_to_bed(reader, 25, offset=1000))
        self.assertListEqual(blocks2, CROSSMAP2)

    def test_fa_to_bed_throws_expected_error(self):
        # test with and without offsets
        # start with unsorted FASTA block

        # grab chrA only and randomize order
        reads = cStringIO.StringIO(SHORT_FASTA_KMERS).read().split("\n")
        reads = reads[40:50] + reads[30:40] + reads[0:16] + reads[20:30]
        reads = "\n".join(reads)
        reader = cStringIO.StringIO(reads)

        # fa_to_bed is a generator, so we need to create a callable to make
        # it a list for it to actually raise the error
        tfunc = lambda x, y, z: list(fa_to_bed(x, y, offset=z))

        self.assertRaises(MalformedFileError, tfunc, reader, 25, 0)

        reader = cStringIO.StringIO(reads)
        self.assertRaises(MalformedFileError, tfunc, reader, 25, 1000)
        reader = cStringIO.StringIO(reads)
        self.assertRaises(MalformedFileError, tfunc, reader, 15, 0)


#===============================================================================
# INDEX: test data
#===============================================================================

SHORT_FASTA = """>chr50a
CCGTGATATGACCTAGGTCGAGAGCTAAGCCTCAATGATGCGCTGGCGAT
>chr30b
CCCTCCTTCCGCTGGCCCCGACTGCCCCAG"""

SHORT_FASTA_KMERS = """>chr50a:0(+)
CCGTGATATGACCTAGGTCGAGAGC
>chr50a:1(+)
CGTGATATGACCTAGGTCGAGAGCT
>chr50a:2(+)
GTGATATGACCTAGGTCGAGAGCTA
>chr50a:3(+)
TGATATGACCTAGGTCGAGAGCTAA
>chr50a:4(+)
GATATGACCTAGGTCGAGAGCTAAG
>chr50a:5(+)
ATATGACCTAGGTCGAGAGCTAAGC
>chr50a:6(+)
TATGACCTAGGTCGAGAGCTAAGCC
>chr50a:7(+)
ATGACCTAGGTCGAGAGCTAAGCCT
>chr50a:8(+)
TGACCTAGGTCGAGAGCTAAGCCTC
>chr50a:9(+)
GACCTAGGTCGAGAGCTAAGCCTCA
>chr50a:10(+)
ACCTAGGTCGAGAGCTAAGCCTCAA
>chr50a:11(+)
CCTAGGTCGAGAGCTAAGCCTCAAT
>chr50a:12(+)
CTAGGTCGAGAGCTAAGCCTCAATG
>chr50a:13(+)
TAGGTCGAGAGCTAAGCCTCAATGA
>chr50a:14(+)
AGGTCGAGAGCTAAGCCTCAATGAT
>chr50a:15(+)
GGTCGAGAGCTAAGCCTCAATGATG
>chr50a:16(+)
GTCGAGAGCTAAGCCTCAATGATGC
>chr50a:17(+)
TCGAGAGCTAAGCCTCAATGATGCG
>chr50a:18(+)
CGAGAGCTAAGCCTCAATGATGCGC
>chr50a:19(+)
GAGAGCTAAGCCTCAATGATGCGCT
>chr50a:20(+)
AGAGCTAAGCCTCAATGATGCGCTG
>chr50a:21(+)
GAGCTAAGCCTCAATGATGCGCTGG
>chr50a:22(+)
AGCTAAGCCTCAATGATGCGCTGGC
>chr50a:23(+)
GCTAAGCCTCAATGATGCGCTGGCG
>chr50a:24(+)
CTAAGCCTCAATGATGCGCTGGCGA
>chr50a:25(+)
TAAGCCTCAATGATGCGCTGGCGAT
>chr30b:0(+)
CCCTCCTTCCGCTGGCCCCGACTGC
>chr30b:1(+)
CCTCCTTCCGCTGGCCCCGACTGCC
>chr30b:2(+)
CTCCTTCCGCTGGCCCCGACTGCCC
>chr30b:3(+)
TCCTTCCGCTGGCCCCGACTGCCCC
>chr30b:4(+)
CCTTCCGCTGGCCCCGACTGCCCCA
>chr30b:5(+)
CTTCCGCTGGCCCCGACTGCCCCAG
"""

CROSSMAP1 = [
    (
        SegmentChain(GenomicSegment("chr50a", 1, 10, "+")),
        SegmentChain(GenomicSegment("chr50a", 1 + 25 - 1, 10 + 25 - 1, "-")),
    ), (
        SegmentChain(GenomicSegment("chr50a", 19, 26, "+")),
        SegmentChain(GenomicSegment("chr50a", 19 + 25 - 1, 26 + 25 - 1, "-")),
    ), (
        SegmentChain(GenomicSegment("chr30b", 0, 6, "+")),
        SegmentChain(GenomicSegment("chr30b", 0 + 25 - 1, 6 + 25 - 1, "-")),
    )
]

CROSSMAP2 = [
    (
        SegmentChain(GenomicSegment("chr50a", 1 + 1000, 10 + 1000, "+")),
        SegmentChain(GenomicSegment("chr50a", 1 + 25 - 1 - 1000, 10 + 25 - 1 - 1000, "-")),
    ), (
        SegmentChain(GenomicSegment("chr50a", 19 + 1000, 26 + 1000, "+")),
        SegmentChain(GenomicSegment("chr50a", 19 + 25 - 1 - 1000, 26 + 25 - 1 - 1000, "-")),
    ), (
        SegmentChain(GenomicSegment("chr30b", 0 + 1000, 6 + 1000, "+")),
        SegmentChain(GenomicSegment("chr30b", 0 + 25 - 1 - 1000, 6 + 25 - 1 - 1000, "-")),
    )
]

# two blocks on chrA, one beginning at position 1, one beginning internally
# one block on chrB, beginning at position 0
CROSSMAP_BLOCK = """>chr50a:1(+)
CGTGATATGACCTAGGTCGAGAGCT
>chr50a:2(+)
GTGATATGACCTAGGTCGAGAGCTA
>chr50a:3(+)
TGATATGACCTAGGTCGAGAGCTAA
>chr50a:4(+)
GATATGACCTAGGTCGAGAGCTAAG
>chr50a:5(+)
ATATGACCTAGGTCGAGAGCTAAGC
>chr50a:6(+)
TATGACCTAGGTCGAGAGCTAAGCC
>chr50a:7(+)
ATGACCTAGGTCGAGAGCTAAGCCT
>chr50a:8(+)
TGACCTAGGTCGAGAGCTAAGCCTC
>chr50a:9(+)
GACCTAGGTCGAGAGCTAAGCCTCA
>chr50a:19(+)
GAGAGCTAAGCCTCAATGATGCGCT
>chr50a:20(+)
AGAGCTAAGCCTCAATGATGCGCTG
>chr50a:21(+)
GAGCTAAGCCTCAATGATGCGCTGG
>chr50a:22(+)
AGCTAAGCCTCAATGATGCGCTGGC
>chr50a:23(+)
GCTAAGCCTCAATGATGCGCTGGCG
>chr50a:24(+)
CTAAGCCTCAATGATGCGCTGGCGA
>chr50a:25(+)
TAAGCCTCAATGATGCGCTGGCGAT
>chr30b:0(+)
CCCTCCTTCCGCTGGCCCCGACTGC
>chr30b:1(+)
CCTCCTTCCGCTGGCCCCGACTGCC
>chr30b:2(+)
CTCCTTCCGCTGGCCCCGACTGCCC
>chr30b:3(+)
TCCTTCCGCTGGCCCCGACTGCCCC
>chr30b:4(+)
CCTTCCGCTGGCCCCGACTGCCCCA
>chr30b:5(+)
CTTCCGCTGGCCCCGACTGCCCCAG
"""



#!/usr/bin/env python
import numpy
import pysam
import warnings
from pkg_resources import resource_filename
from nose.tools import assert_true, assert_equal, assert_greater_equal
from plastid.genomics.map_factories import FivePrimeMapFactory,\
                                       ThreePrimeMapFactory,\
                                       CenterMapFactory,\
                                       VariableFivePrimeMapFactory,\
                                       StratifiedVariableFivePrimeMapFactory
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from plastid.genomics.genome_array import BAMGenomeArray
from plastid.util.services.mini2to3 import cStringIO


class TestBAM_MappingRules(object):
    @classmethod
    def setUpClass(cls):
        min_ = 25
        max_ = 40
        cls.strands = ("+", "-")
        cls.segs = {X: GenomicSegment("mock", 0, 2000, X) for X in cls.strands}

        cls.reads = {
            Y: [cls.make_alignment(0, X, Y) for X in range(min_, max_)]
            for Y in ("+", "-")
        }
        cls.expected = {}
        for mapping in ("fiveprime", "threeprime", "center"):
            for param in (0, 10):
                for strand in cls.strands:
                    cls.expected[(mapping, param, strand)] = numpy.zeros(2000)

        cls.expected[("fiveprime", 0, "+")][0] = max_ - min_
        cls.expected[("fiveprime", 10, "+")][10] = max_ - min_
        cls.expected[("fiveprime", 0, "-")][min_ - 1:max_ - 1] = 1
        cls.expected[("fiveprime", 10, "-")][min_ - 11:max_ - 11] = 1

        cls.expected[("threeprime", 0, "-")][0] = max_ - min_
        cls.expected[("threeprime", 10, "-")][10] = max_ - min_
        cls.expected[("threeprime", 0, "+")][min_ - 1:max_ - 1] = 1
        cls.expected[("threeprime", 10, "+")][min_ - 11:max_ - 11] = 1

        for my_len in range(min_, max_):
            cls.expected[("center", 0, "+")][:my_len] += 1.0 / my_len
            cls.expected[("center", 0, "-")][:my_len] += 1.0 / my_len
            cls.expected[("center", 10, "+")][10:my_len - 10] += 1.0 / (my_len - 2 * 10)
            cls.expected[("center", 10, "-")][10:my_len - 10] += 1.0 / (my_len - 2 * 10)

        cls.map_factories = {
            "fiveprime": FivePrimeMapFactory,
            "threeprime": ThreePrimeMapFactory,
            "fiveprime_variable": VariableFivePrimeMapFactory,
            "center": CenterMapFactory
        }

    @staticmethod
    def make_alignment(start_pos, end_pos, strand):
        read = pysam.AlignedSegment()
        read.reference_start = start_pos
        my_length = end_pos - start_pos
        read.query_sequence = "N" * my_length
        read.reference_id = 0
        read.is_reverse = True if strand == "-" else False
        read.cigarstring = "%sM" % my_length
        return read

    def check_mapping(self, map_name, map_param, strand):
        fn = self.map_factories[map_name](map_param)
        expected = self.expected[(map_name, map_param, strand)]
        reads_out, count_array = self.map_factories[map_name](map_param)(
            self.reads[strand], self.segs[strand]
        )
        msg1 = "failed to return reads with %s mapping with param %s on strand '%s'." % (
            map_name, map_param, strand
        )
        msg2 = "failed %s mapping vector with param %s on strand '%s'." % (
            map_name, map_param, strand
        )
        assert_equal(reads_out, self.reads[strand], msg1)
        assert_true((count_array == expected).all(), msg2)

    def test_fiveprime_threeprime_center(self):
        for mapping in ("fiveprime", "threeprime", "center"):
            for offset in 0, 10:
                for strand in self.strands:
                    yield self.check_mapping, mapping, offset, strand

    def test_fiveprime_variable(self):
        offset_dicts = {}
        fancy = {X: X // 2 for X in range(25, 40)}
        fancy_reads = {X: numpy.zeros(2000) for X in self.strands}
        for read in self.reads["+"]:
            fancy_reads["+"][read.positions[fancy[len(read.positions)]]] += 1
        for read in self.reads["-"]:
            fancy_reads["-"][read.positions[-fancy[len(read.positions)] - 1]] += 1

        fancy_plus = numpy.zeros(2000)
        fancy_minus = numpy.zeros(2000)

        offset_dicts[("default_only", "+")] = ({"default": 0}, self.expected[("fiveprime", 0, "+")])
        offset_dicts[("default_only", "-")] = ({"default": 0}, self.expected[("fiveprime", 0, "-")])
        offset_dicts[("fancy", "+")] = (fancy, fancy_reads["+"])
        offset_dicts[("fancy", "-")] = (fancy, fancy_reads["-"])

        for (name, strand), (dict_, expected) in offset_dicts.items():
            fn = VariableFivePrimeMapFactory(dict_)
            reads_out, count_array = fn(self.reads[strand], self.segs[strand])
            msg = "Failed fiveprime variable mapping for test %s(%s)" % (name, strand)
            assert_true((count_array == expected).all(), msg)

    def test_fiveprime_variable_from_file(self):
        offset_dicts = {}
        fancy = {X: X // 2 for X in range(25, 40)}
        fancy_reads = {X: numpy.zeros(2000) for X in self.strands}
        for read in self.reads["+"]:
            fancy_reads["+"][read.positions[fancy[len(read.positions)]]] += 1
        for read in self.reads["-"]:
            fancy_reads["-"][read.positions[-fancy[len(read.positions)] - 1]] += 1

        fancy_plus = numpy.zeros(2000)
        fancy_minus = numpy.zeros(2000)

        offset_dicts[("default_only", "+")] = ({"default": 0}, self.expected[("fiveprime", 0, "+")])
        offset_dicts[("default_only", "-")] = ({"default": 0}, self.expected[("fiveprime", 0, "-")])
        offset_dicts[("fancy", "+")] = (fancy, fancy_reads["+"])
        offset_dicts[("fancy", "-")] = (fancy, fancy_reads["-"])

        for (name, strand), (dict_, expected) in offset_dicts.items():
            fhlike = cStringIO.StringIO("\n".join(["%s\t%s" % (K, V) for K, V in dict_.items()]))
            fn = VariableFivePrimeMapFactory.from_file(fhlike)
            reads_out, count_array = fn(self.reads[strand], self.segs[strand])
            msg = "Failed fiveprime variable mapping for test %s(%s)" % (name, strand)
            assert_true((count_array == expected).all(), msg)

    def check_unmappable_raises_warnings(self, test_name, map_param, strand):
        map_factory = self.map_factories[test_name]
        msg = "Map function '%s' failed to raise warning for unmappable reads on strand '%s'" % (
            test_name, strand
        )
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter("always")
            fn = map_factory(map_param)
            reads_out, count_array = fn(self.reads[strand], self.segs[strand])

        assert_greater_equal(len(warns), 0, msg)

    def test_unmappable_raises_warnings(self):
        func_params = {
            "fiveprime": 30,
            "threeprime": 30,
            "center": 15,
            "fiveprime_variable": {
                25: 10,
                "default": 28
            }
        }
        for test_name, map_param in func_params.items():
            for strand in self.strands:
                yield self.check_unmappable_raises_warnings, test_name, map_param, strand

    def check_unmappable_not_mapped_in_vector_or_returned(
            self, test_name, map_param, expected_num_reads, strand
    ):
        msg1 = "Map function '%s' failed to filter out short reads on strand '%s' from returned reads" % (
            test_name, strand
        )
        msg2 = "Map function '%s' failed to filter out short reads on strand '%s' from count vector" % (
            test_name, strand
        )
        fn = self.map_factories[test_name](map_param)
        reads_out, count_array = fn(self.reads[strand], self.segs[strand])
        assert_equal(len(reads_out), expected_num_reads, msg1)
        assert_equal(count_array.sum(), expected_num_reads, msg2)

    def test_unmappable_not_mapped_in_vector_or_returned(self):
        func_params = {
            "fiveprime"  : 30,
            "threeprime" : 30,
            "center"     : 15,
            "fiveprime_variable": {
                25       : 10,
                "default": 28
            } # yapf: disable
        }
        expected_num = {
            "fiveprime":
            len([X for X in self.reads["+"] if len(X.positions) > func_params["fiveprime"]]),
            "threeprime":
            len([X for X in self.reads["+"] if len(X.positions) > func_params["threeprime"]]),
            "center":
            len([X for X in self.reads["+"] if len(X.positions) > 2 * func_params["center"]]),
            "fiveprime_variable":
            len([X for X in self.reads["+"] if len(X.positions) > 28 or len(X.positions) == 25]),
        }
        for test_name, map_param in func_params.items():
            expected_num_reads = expected_num[test_name]
            for strand in self.strands:
                yield self.check_unmappable_not_mapped_in_vector_or_returned, test_name, map_param, expected_num_reads, strand

    def test_variable_stratified_mapping_plus(self):
        offsets = {26: 6, 27: 22, 28: 13, 29: 4, 30: 5}

        chains = {
            "fw":
            SegmentChain(
                GenomicSegment('chrII', 392959, 393180, '+'),
                GenomicSegment('chrII', 393510, 394742, '+'),
                GenomicSegment('chrII', 394860, 394901, '+'),
                ID='YBR078W_mRNA'
            ),
            "rc":
            SegmentChain(
                GenomicSegment('chrVIII', 189061, 189749, '-'),
                GenomicSegment('chrVIII', 189850, 190017, '-'),
                ID='YHR041C_mRNA'
            )
        }
        expected = {
            "fw":
            numpy.loadtxt(
                resource_filename("plastid", "test/data/stratmap/strat_fw_vec.txt"), delimiter="\t"
            ),
            "rc":
            numpy.loadtxt(
                resource_filename("plastid", "test/data/stratmap/strat_rc_vec.txt"), delimiter="\t"
            ),
        }
        ga = BAMGenomeArray([resource_filename("plastid", "test/data/stratmap/strat.bam")])
        ga.set_mapping(StratifiedVariableFivePrimeMapFactory(offsets, 26, 30))

    def test_variable_stratified_mapping_minus(self):
        pass




#===============================================================================
# Imports
#===============================================================================

from plastid.genomics.c_common cimport (
    Strand,
    forward_strand,
    reverse_strand,
    unstranded,
    undef_strand,
    true,
    false,
    ExBool,
    bool_exception,
    strand_to_str,
    str_to_strand,
)

from cpython cimport array

cimport numpy
import numpy
import array



#===============================================================================
# Functions, forward declarations, objects
#===============================================================================

cdef class GenomicSegment(object)
cdef class SegmentChain(object)
cdef class Transcript(SegmentChain)


# GenomicSegment used as a signal for/by various readers
cdef GenomicSegment NullSegment

# functions for converting coordinates
cpdef list positions_to_segments(str, str, object)
cpdef list positionlist_to_segments(str, str, list)
cpdef Transcript add_three_for_stop_codon(Transcript)
cpdef list merge_segments(list)

# Various internals used by SegmentChain/Transcript
cdef void nonecheck(object, str, str)
cdef bint check_segments(SegmentChain, tuple) except False
cdef ExBool chain_richcmp(SegmentChain, SegmentChain, int) except bool_exception
cdef ExBool transcript_richcmp(Transcript, Transcript, int) except bool_exception


# BED parsing helpers
cdef dict get_attr_from_bed_column_tuples(list, int, int, list)
cdef dict get_attr_from_bed_column_names(list, int, list)
cdef dict get_attr_from_bed_column_number(list, int, int)
cdef dict get_standard_bed_attr(list, int)
cdef dict get_attr_from_bed(str line, object extra_columns=*)
cdef list get_segments_from_bed(dict attr)


#===============================================================================
# Extension types / classes
#===============================================================================


# Single, continuous stretches of the genome
cdef class GenomicSegment(object):                      # 20 bytes + str
    cdef str chrom
    cdef long start                                     # 8 bytes
    cdef long end                                       # 8 bytes
    cdef Strand c_strand                                # 4 bytes if int
    cpdef bint contains(self, GenomicSegment)
    cpdef bint overlaps(self, GenomicSegment)
    cpdef bint _cmp_helper(self, GenomicSegment, int)
    cpdef str as_igv_str(self)

# Potentially discontinuous entities made up of zero or more GenomicSegments
# can have arbitrary attributes
cdef class SegmentChain(object):                       # >= 836 bytes
    cdef:
        list _segments, _mask_segments                 # 72 bytes/each -> 144 + contents
        readonly GenomicSegment spanning_segment       # 20 bytes + str
        readonly long length                           # 8 bytes
        readonly long masked_length                    # 8 bytes

        array.array _position_hash                     # 56 + contents
        array.array _position_mask # really should be bint, but we can't use it there
        array.array _endmap
        public dict attr                               # 280 bytes + contents
        #dict _inverse_hash                             # 280 bytes + contents

    # maintenance of chain internals
    cdef bint c_add_segments(self, tuple) except False
    cdef bint _set_segments(self, list) except False
    cdef array.array _get_position_hash(self)
    cdef bint _set_masks(self, list) except False
    cdef void c_reset_masks(self)
    #cdef dict _get_inverse_hash(self)

    # comparison operators
    cdef  list c_shares_segments_with(self, SegmentChain)
    cdef  ExBool c_covers(self , SegmentChain) except bool_exception
    cdef  ExBool c_antisense_overlaps(self, SegmentChain) except bool_exception
    cdef  ExBool c_unstranded_overlaps(self, SegmentChain) except bool_exception
    cdef  ExBool c_overlaps(self, SegmentChain) except bool_exception
    cdef  ExBool c_contains(self, SegmentChain) except bool_exception

    # position info - c impl
    cdef list c_get_position_list(self)
    cdef set c_get_position_set(self)
    cdef numpy.ndarray c_get_position_array(self,bint)
    cpdef set get_masked_position_set(self)

    # subchains/coordinates
    cdef long c_get_genomic_coordinate(self, long, bint) except -1
    cdef long c_get_segmentchain_coordinate(self, long, bint) except -1
    cdef SegmentChain c_get_subchain(self, long, long, bint)

    # strand changes
    cdef SegmentChain _restrand(self, Strand strand)#,extra_attr=*)
    cpdef SegmentChain get_unstranded(self)#, extra_attr=*)
    cpdef SegmentChain get_antisense(self)#, extra_attr=*)



# Specialized SegmentChain for Transcripts
cdef class Transcript(SegmentChain):
    cdef:
        object cds_genome_start, cds_genome_end, cds_start, cds_end

    cdef bint _update_cds(self) except False
    cdef bint _update_from_cds_start(self) except False
    cdef bint _update_from_cds_end(self) except False
    cdef bint _update_from_cds_genome_start(self) except False
    cdef bint _update_from_cds_genome_end(self) except False



#!/usr/bin/env python
from plastid.util.io.openers import (get_short_name, pretty_print_dict)
from nose.tools import assert_equal


def test_get_short_name():
    tests = [
        ("test", "test",{}),
        ("test.py", "test", dict(terminator=".py")),
        ("/home/jdoe/test.py", "test", dict(terminator=".py")),
        ("/home/jdoe/test.py.py","test.py", dict(terminator=".py")),
        ("/home/jdoe/test.py.2", "test.py.2", {}),
        ("/home/jdoe/test.py.2", "test.py.2", dict(terminator=".py")),
        ("plastid.bin.test", "test", dict(separator="\.", terminator=""))
    ]
    for inp, expected, kwargs in tests:
        found = get_short_name(inp, **kwargs)
        msg = "test_get_short_name(): failed on input '%s'. Expected '%s'. Got '%s'" % (
            inp, expected, found
        )
        yield assert_equal, expected, found, msg


def test_pretty_print_dict():
    dtmp = {
        "a": 1,
        "b": 2.3,
        "c": "some string",
        "d": "some string with 'subquotes' inside",
        "e": (3, 4, 5),
        "somereallyreallylongname": "short val",
    }
    expected = """{
          'a'                        : 1,
          'b'                        : 2.3,
          'c'                        : 'some string',
          'd'                        : 'some string with 'subquotes' inside',
          'e'                        : (3, 4, 5),
          'somereallyreallylongname' : 'short val',
}
"""
    found = pretty_print_dict(dtmp)
    assert_equal(
        expected, found,
        "Dictionary did not pretty-print!\nExpected:\n%s\n\nFound:\n%s\n\n" % (expected, found)
    )



#!/usr/bin/env python
"""This subpackage contains plotting utilities to build some compound plots
not found in matplotlib. For demos, see :doc:`/examples/z_plotting`.


    ======================================    =========================================================
    **Submodule**                             **Contents**
    --------------------------------------    ---------------------------------------------------------
    :mod:`plastid.plotting.colors`            Utilities for manipulating, converting, darkening,
                                              and lightening colors

    :mod:`plastid.plotting.plots`             Plots, such as heatmaps with summary profiles,
                                              scatter plots with kernel density estimates of 
                                              marginal distributions, MA plots, et c

    :mod:`plastid.plotting.plotutils`         Utility functions for preprocessing data or manipulating
                                              axes
    ======================================    =========================================================
"""
from plastid.plotting.plots import scatterhist_x, scatterhist_y, scatterhist_xy, \
                                   ma_plot, phase_plot, \
                                   profile_heatmap, \
                                   triangle_plot, \
                                   stacked_bar, \
                                   kde_plot



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.findjuncs`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES
from plastid.bin.findjuncs import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.findjuncs",
    "ref_file_path": resource_filename("plastid", "test/data/"),
    "temp_file_path": tempfile.mkdtemp(prefix="findjuncs"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
basename = os.path.join(test_info["temp_file_path"], "test_findjuncs")

findjuncs_tests = [
    # find all junctions in one GTF
    (
        "%s --annotation_files %s --annotation_format GTF2" % (basename, REF_FILES["yeast_gtf"]),
        [REF_FILES["yeast_findjuncs_bed"]],
        ["%s.bed" % basename], ["--no_header --sort_keys 3"]
    ),

    # find all junctions in two GTFs; report only once
    (
        "%s --annotation_files %s %s" % (basename, REF_FILES["yeast_gtf"], REF_FILES["yeast_gtf"]),
        [REF_FILES["yeast_findjuncs_bed"]],
        ["%s.bed" % basename], ["--no_header --sort_keys 3"]
    ),

    # find all junctions in one GTF, export tophat juncs
    (
        "%s --annotation_files %s --export_tophat" % (basename, REF_FILES["yeast_gtf"]), [
            REF_FILES["yeast_findjuncs_bed"],
            REF_FILES["yeast_findjuncs_juncs"],
        ],
        ["%s.bed" % basename, "%s.juncs" % basename], ["--no_header --sort_keys 3", "--no_header"]
    ),

    # pool different but overlapping sets of junctions from two BED files, export tophat juncs
    (
        "%s --annotation_files %s %s --annotation_format BED --export_tophat" %
        (basename, REF_FILES["yeast_findjuncs_top"], REF_FILES["yeast_findjuncs_bot"]),
        [REF_FILES["yeast_findjuncs_bed"], REF_FILES["yeast_findjuncs_juncs"]],
        ["%s.bed" % basename, "%s.juncs" % basename], ["--no_header --sort_keys 3", "--no_header"]
    ),
]
"""Functional tests of :py:mod:`plastid.bin.findjuncs`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod`plastid.bin.findjuncs`"""
    for x in execute_helper(test_info, findjuncs_tests):
        yield x



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.make_wiggle`"""
import tempfile
import shutil
import os
import shlex
import nose

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import (
    RPATH,
    REF_FILES,
    COUNT_OPTIONS,
    ANNOTATION_OPTIONS,
    MASK_OPTIONS,
)
from plastid.genomics.genome_array import GenomeArray
from plastid.bin.make_wiggle import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants used by tests
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.make_wiggle",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="make_wiggle"),
}
"""Constants used in tests below"""

#===============================================================================
# INDEX: programmatically define tests
#===============================================================================

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
make_wiggle_tests = []
samples = {
    "fiveprime_variable": "--fiveprime_variable --offset %s" % REF_FILES["yeast_psite"],
}

for side in ("fiveprime", "threeprime"):
    for offset in (0, 15):
        samples["%s_%s" % (side, offset)] = "--%s --offset %s" % (side, offset)

for nibble in (0, 12):
    samples["center_%s" % nibble] = "--center --nibble %s" % nibble

for k, v in samples.items():
    fbase = os.path.join(test_info["temp_file_path"], "test_make_wiggle_%s" % k)
    make_wiggle_tests.append(
        (
            "%s --count_files %s -o %s " % (v, REF_FILES["yeast_rp_bam"], fbase), [
                os.path.join(test_info["ref_file_path"], "gen_reads_%s_fw.wig" % k),
                os.path.join(test_info["ref_file_path"], "gen_reads_%s_rc.wig" % k),
            ], [
                os.path.join(test_info["temp_file_path"], "test_make_wiggle_%s_fw.wig" % k),
                os.path.join(test_info["temp_file_path"], "test_make_wiggle_%s_rc.wig" % k),
            ]
        )
    )

#===============================================================================
# INDEX: test functions and helpers
#===============================================================================


def do_check(argstr, ref_files, test_files):
    """Check that test wiggle files match reference files

    Parameters
    ----------
    ref_files : tuple
        Tuple of filenames corresponding to forward and reverse strand wiggles
        for reference dataset

    test_files : list
        Tuple of filenames corresponding to forward and reverse strand wiggles
        for test dataset
    """
    err_message = "Unequal output in files %s vs %s, module %s" % (
        ref_files, test_files, test_info["module_name"]
    )
    test_info["test_method"](shlex.split(argstr))
    ref_ga = GenomeArray()
    test_ga = GenomeArray()
    with open(ref_files[0]) as fh:
        ref_ga.add_from_wiggle(fh, "+")
    with open(ref_files[1]) as fh:
        ref_ga.add_from_wiggle(fh, "-")
    with open(test_files[0]) as fh:
        test_ga.add_from_wiggle(fh, "+")
    with open(test_files[1]) as fh:
        test_ga.add_from_wiggle(fh, "-")
    nose.tools.assert_true(ref_ga.__eq__(test_ga, tol=1e-8), err_message)


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Execute functional test for :py:mod:`plastid.bin.make_wiggle`"""
    for argstr, ref_files, test_files in make_wiggle_tests:
        yield do_check, argstr, ref_files, test_files


def tearDownModule():
    """Remove test dataset files after unit tests are complete"""
    if test_info["temp_file_path"] != "":
        shutil.rmtree(test_info["temp_file_path"])

    cleanup_resources()



#!/usr/bin/env python
"""This module contains custom exception and warning classes, implements 
a custom warning filter action, called `"onceperfamily"`, and monkey-patches
warning output to improve legibility.

Contents:

.. contents::
   :local:

The `onceperfamily` action
--------------------------
`onceperfamily` groups warning messages by families regular expressions,
and only prints the first warning instance that matches a given family's 
regular expression. In contrast, Python's native `once` action prints any string
literal once, even if it matches the same regex as another warning already given.

To use this action, use the following two functions:

  - :func:`filterwarnings` to create the warnings filter. Because :func:`filterwarnings`
    wraps Python's :func:`warnings.filterwarnings`, it may be used as a drop-in
    replacement for creation of any warnings filter.
    
  - :func:`warn` or :func:`warn_explicit`. Again, these are drop-in replacements
    for Python's :func:`warnings.warn` and :func:`warnings.warn_explicit` that
    additionally check the `onceperfamily` filters.
    
For convenience, there are also functions that both issue a warning, and create
a `onceperfamily` filter for it if one doesn't already exist:

  - :func:`warn_onceperfamily`
  - :func:`warn_explicit_onceperfamily`


Exception types
---------------
|MalformedFileError|
    Raised when a file cannot be parsed as expected, and
    execution must halt


Warning types
-------------
|ArgumentWarning|
    Warning for command-line arguments that:
    
      - are nonsenical, but recoverable
      - together might cause very slow execution
        (e.g. run would be optimized by other combinations)

|FileFormatWarning|
    Warning for slightly malformed but usable files

|DataWarning|
    Warning raised when:

      - data has unexpected attributes
      - data has nonsensical, but recoverable values for attributes
      - when values are out of the domain of a given operation,
        but skipping the operation or estimating the value is permissible


See also
--------
:mod:`warnings`
    Warnings module
"""
import re
import warnings
import inspect
import linecache
import textwrap
from plastid.util.io.filters import colored

_wrapper = textwrap.TextWrapper(break_long_words=False, width=77)

#===============================================================================
# INDEX: Warning and exception classes
#===============================================================================


class MalformedFileError(Exception):
    """Exception class for when files cannot be parsed as they should be
    """

    def __init__(self, filename, message, line_num=None):
        """Create a |MalformedFileError|
        
        Parameters
        ----------
        filename : str
            Name of file causing problem
        
        message : str
            Message explaining how the file is malformed.
        
        line_num : int or None, optional
            Number of line causing problems
        """
        self.filename = filename
        self.msg = message
        self.line_num = line_num

    def __str__(self):
        if self.line is None:
            return "Error opening file '%s': %s" % (self.filename, self.msg)
        else:
            return "Error opening file '%s' at line %s: %s" % (
                self.filename, self.line_num, self.msg
            )


class ArgumentWarning(Warning):
    """Warning for nonsensical but recoverable combinations of command-line arguments,
    or arguments that risk slow program execution"""
    pass


class FileFormatWarning(Warning):
    """Warning for slightly malformed but usable files"""
    pass


class DataWarning(Warning):
    """Warning for unexpected attributes of data.
    Raised when:

      - data has unexpected attributes
      - data has nonsensical, but recoverable values
      - values are out of the domain of a given operation, but execution
        can continue if the value is estimated or the operation skipped
    """


#===============================================================================
# INDEX: Plastid's extensions to Python warnings
#===============================================================================

pl_once_registry = {}
"""Registry of `onceperfamily` warnings that have been seen in the current execution context"""

pl_filters = []
"""Plastid's own warnings filters, which allow additional actions compared to Python's"""


def filterwarnings(action, message="", category=Warning, module="", lineno=0, append=0):
    """Insert an entry into the warnings filter. Behaviors are as in :func:`warnings.filterwarnings`,
    except the additional action `'onceperfamily'` can be used to allow one warning per `family`
    of messages, specified by a regex. 
    
    This allows individual warnings to give more detailed information, without each being
    regarded as its own warning by Python's warning system (the defualt behavior of `'once'`).
    
    
    Parameters
    ----------
    action : str
        How the warning should be filtered. Accceptable values are "error",
        "ignore", "always", "default", 'module", "once", and "onceperfamily"
        
    message : str, optional
        str that can be compiled to a regex, used to detect warnings. If "onceperfamily"
        is chosen, only the first warning to give a string that matches the regex
        will be shown. For other actions, behaviors are as described in :mod:`warnings`.
        (Default: `""`, match any message)
        
    category : Warning or subclass, optional
        Type of warning. (Default: :class:`Warning`)

    module : str, optional
        str that can be compiled to a regex, limiting the warning behavior to modules
        that match that regex. (Default: `""`, match all modules)
        
    lineno : int, optional
        integer line used to specify warning in source code. If 0 (default), match
        all warnings regardless of line number.
        
    append : int, optional
        If 1, add warning to end of filter list. If 0 (default), insert warning at 
        beginning of filters list.
        
    
    See also
    --------
    warnings.filterwarnings
        Python's warnings filter
    """
    tup = (action, re.compile(message, re.I), category, re.compile(module), lineno)
    if action == "onceperfamily":
        if tup in pl_filters:
            return
        else:
            if append == 1:
                pl_filters.append(tup)
            else:
                pl_filters.insert(0, tup)
    else:
        warnings.filterwarnings(
            action, message=message, category=category, module=module, lineno=lineno, append=append
        )


def warn_onceperfamily(message, pattern=None, category=None, stacklevel=1):
    """Issue a warning and create a warning filter for that warning if it does not already exist
    
    Parameters
    ----------
    message : str
        Message of warning. Can be a string that compiles to a regex.
        Printed as warning text and used to create warning filter if `pattern`
        is `None`. 
    
    pattern : str or None, optional
        If not `None`, override message when generating warnings filter
        
    category: :class:`Warning`, or subclass, optional
        Type of warning
        
    stacklevel : int
        Ignored
        
    See also
    --------
    plastid.util.services.exceptions.filterwarnings
        plastid-specific warnings filters
    
    warnings.warn
        Python's warning system, which this wraps
    """
    if pattern is None:
        pattern = message
        filterwarnings("onceperfamily", message=pattern, category=category)

    warn(message, category=category, stacklevel=stacklevel)


def warn_explicit_onceperfamily(
        message,
        category,
        filename,
        lineno,
        pattern=None,
        module=None,
        registry=None,
        module_globals=None
):
    """Low-level interface to issue warnings, allowing `plastid`-specific warnings filters

    Parameters
    ----------
    message : str
        Message
    
    pattern : str or None, optional
        If not `None`, override message when generating warnings filter
        
    category: :class:`Warning`, or subclass, optional
        Type of warning

    filename : str
        Name of module from which warning is issued
    
    lineno : int, optional
        Line in module at which warning is called

    module : str, optional
        Module name
    
    registry : dict, optional
        Registry of ignore filters (see source code for :func:`warnings.warn_explicit`
    
    module_globals : dict, optional
        Dictionary of module-level variables
        
    
    See also
    --------
    plastid.util.services.exceptions.filterwarnings
        plastid-specific warnings filters

    warnings.warn_explicit
        Python's warning system, which this wraps
    """
    if pattern is None:
        pattern = message
        filterwarnings("onceperfamily", message=pattern, category=category)

    if module is None:  # or module_globals is None:
        frame = inspect.currentframe()
        if frame is None:
            module = __name__
        else:
            try:
                module = inspect.getmodule(frame.f_back.f_code).__name__
            finally:
                del frame

    warn_explicit(
        pattern,
        category,
        filename,
        lineno,
        module=module,
        registry=registry,
        module_globals=module_globals
    )


def warn(message, category=None, stacklevel=1):
    """Issue a non-essential warning to users, allowing `plastid`-specific warnings filters
    
    Parameters
    ----------
    message : str
        Message
    
    category: :class:`Warning`, or subclass, optional
        Type of warning
        
    stacklevel : int
        Ignored
        
    See also
    --------
    plastid.util.services.exceptions.filterwarnings
        plastid-specific warnings filters
    
    warnings.warn
        Python's warning system, which this wraps
    """
    if category is None:
        category = UserWarning

    _, filename, lineno, _, _, _ = inspect.stack()[stacklevel]
    warn_explicit(message, category, filename, lineno, module=filename)


def warn_explicit(
        message, category, filename, lineno, module=None, registry=None, module_globals=None
):
    """Low-level interface to issue warnings, allowing `plastid`-specific warnings filters

    Parameters
    ----------
    message : str
        Message
    
    category: :class:`Warning`, or subclass, optional
        Type of warning

    filename : str
        Name of module from which warning is issued
    
    lineno : int, optional
        Line in module at which warning is called

    module : str, optional
        Module name
    
    registry : dict, optional
        Registry of ignore filters (see source code for :func:`warnings.warn_explicit`
    
    module_globals : dict, optional
        Dictionary of module-level variables
        
    
    See also
    --------
    plastid.util.services.exceptions.filterwarnings
        plastid-specific warnings filters

    warnings.warn_explicit
        Python's warning system, which this wraps
    """
    global pl_once_registry
    if module is None:
        frame = inspect.currentframe()
        if frame is None:
            module = __name__
        else:
            try:
                module = inspect.getmodule(frame.f_back.f_code).__name__
            finally:
                del frame

    for _, pat, filter_category, mod, filter_line in pl_filters:
        if pat.match(message) and issubclass(category,filter_category) and\
           (module is None or mod.match(module)) and\
           (filter_line == 0 or filter_line == lineno):

            tup = (pat.pattern, filter_category, mod, filter_line)
            if tup in pl_once_registry:
                return
            else:
                pl_once_registry[tup] = 1
                break

    warnings.warn_explicit(
        message,
        category,
        filename,
        lineno,
        module=module,
        registry=registry,
        module_globals=module_globals
    )


def formatwarning(message, category, filename, lineno, file=None, line=None):
    """Wrapper to colorize warnings for readability. Overrides :func:`warnings.formatwarning`
    
    Parameters
    ----------
    message : str
        Warning message
        
    category : Warning
        Class (not instance) of warning
    
    filename : str
        Name of file calling warning
    
    lineno : int
        Line in file calling warning
    
    file : something implementing a `write` method 
        File to which output is written. Default: :obj:`sys.stderr`
        
    line : str
        Text of line in file calling warning. If `None`, `line` is taken
        to be line number `lineno` of `filename` 
    
    Returns
    -------
    str
        Pretty-printed warning message
    """
    sep = colored("-" * 75, color="cyan")
    message = str(message)
    if not "\n" in message:
        message = _wrapper.fill(str(message))

    message = colored(message, color="white", attrs=["bold"])
    name = colored(category.__name__, color="cyan", attrs=["bold"])

    if line is None:
        numwidth = len(str(lineno + 3))
        fmtstr = "{0: >%ss} {1}" % (numwidth)
        lines = []
        for x in range(max(0, lineno - 2), lineno + 3):
            tmpline = linecache.getline(filename, x).strip("\n")
            if tmpline:
                attrs = ["bold"] if x == lineno else []
                lines.append(
                    fmtstr.format(
                        colored(x, color="green", attrs=attrs), colored(tmpline, attrs=attrs)
                    )
                )
        line = "\n".join(lines)

    filename = "in %s, line %s:" % (colored(filename, color="cyan"), lineno)

    ltmp = [sep, name, message, filename, "", line, "", sep, ""]

    return "\n".join(ltmp)


warnings.formatwarning = formatwarning



#!/usr/bin/env python
"""This script identify all the unique splice junctions in one or more transcript
annotations, and exports these as a `BED`_ file with one splice junction per line
Optionally, this script can also export junctions as a `Tophat`_ ``.juncs`` file.

If a splice junction appears multiple times (e.g. used by more than one transcript),
only the first occurrence of the junction will be reported. Scores, if present,
are exported unaltered in `BED`_ output.

Examples:


.. code-block:: shell

   # identify splice junctions from a transcript annotation supplied in GTF2
   # creates output file 'annotation.bed'
   $ findjuncs my_annotation --annotation_format GTF2 \\
               --annotation_files transcripts.gtf
    
   # merge unique annotations from annotation.bed and newly_discovered.bed,
   # export only unique junctions to 'merged_unique.bed'
   $ findjuncs merged_unique --annotation_format BED \\
               --annotation_files annotation.bed newly_discovered.bed


See also
--------
:mod:`plastid.bin.slidejuncs`
   Script that makes richer comparisons between discovered and annotated
   junctions, using genomic sequence and :py:mod:`plastid.bin.crossmap`
   results to classify junctions
"""
"""
From the Tophat specification for `.juncs` files:

    Junctions are specified one per line, in a tab-delimited format. Records
    look like::
    
        <chrom> <left> <right> <+/->
    
    left and right are zero-based coordinates, and specify the last character of
    the left sequenced to be spliced to the first character of the right sequence,
    inclusive. That is, the last and the first positions of the flanking exons.
    Users can convert junctions.bed (one of the TopHat outputs) to this format
    using bed_to_juncs < junctions.bed > new_list.juncs where bed_to_juncs can
    be found under the same folder as tophat

See http://ccb.jhu.edu/software/tophat/index.shtml for more information.
"""
import sys
import argparse
import inspect
import warnings
from plastid.genomics.roitools import SegmentChain
from plastid.util.scriptlib.argparsers import AnnotationParser, BaseParser
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import argsopener, get_short_name
from plastid.util.scriptlib.help_formatters import format_module_docstring

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

_ANNOTATION_INPUT_CHOICES = ["BED", "BigBed", "GTF2", "GFF3", "PSL"]
_ANNOTATION_DISABLED = ["add_three", "annotation_file"]


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: sys.argv[1:] (actually command-line arguments)
    """
    ap = AnnotationParser(input_choices=_ANNOTATION_INPUT_CHOICES)
    annotation_file_parser = ap.get_parser()

    bp = BaseParser()
    base_parser = bp.get_parser()

    # yapf: disable
    parser = argparse.ArgumentParser(
        description     = format_module_docstring(__doc__),
        formatter_class = argparse.RawDescriptionHelpFormatter,
        parents         =[base_parser, annotation_file_parser]
    )
    parser.add_argument(
        "--export_tophat",
        default =False,
        action  ="store_true",
        help    ="Export tophat `.juncs` file in addition to BED output"
    )
    parser.add_argument("outbase", type=str, help="Basename for output files")
    # yapf: enable

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    transcripts = ap.get_transcripts_from_args(args, printer=printer, return_type=SegmentChain)

    with argsopener("%s.bed" % args.outbase, args, "w") as bed_out:
        if args.export_tophat == True:
            tophat_out = open("%s.juncs" % args.outbase, "w")

        printer.write("params: " + " ".join(argv))
        printer.write("Detecting & comparing junctions...")
        ex_pairs = {}

        c = 0
        u = 0
        for chain in transcripts:
            if len(chain) > 1:  # if multi-exon
                chrom = chain.chrom
                strand = chain.strand
                try:
                    ep = ex_pairs[(chrom, strand)]
                except KeyError:
                    ex_pairs[(chrom, strand)] = []
                    ep = ex_pairs[(chrom, strand)]

                for i in range(0, len(chain) - 1):

                    seg1 = chain[i]
                    seg2 = chain[i + 1]
                    if c % 1000 == 0 and c > 0:
                        printer.write("Processed %s junctions. Found %s unique..." % (c, u))
                    c += 1
                    key = (seg1.end, seg2.start)

                    if key not in ep:
                        ep.append(key)
                        u += 1
                        new_chain = SegmentChain(seg1, seg2)
                        bed_out.write(new_chain.as_bed())
                        if args.export_tophat == True:
                            my_junc = (chrom, seg1.end - 1, seg2.start, strand)
                            tophat_out.write("%s\t%s\t%s\t%s\n" % my_junc)

                        del new_chain

                    del seg1
                    del seg2

            del chain

        printer.write("Processed %s total junctions. Found %s unique." % (c, u))

        bed_out.close()
        if args.export_tophat == True:
            tophat_out.close()

    printer.write("Done.")


if __name__ == "__main__":
    main()



"""Base class for `BigBed`_ and `BigWig`_ readers, both of which are implemented
as a Python binding for the C library of Jim Kent's utilties for the UCSC genome
browser.


See also
--------
`Kent2010 <http://dx.doi.org/10.1093/bioinformatics/btq351>`_
    Description of BigBed and BigWig formats. Especially see supplemental data.

`Source repository for Kent utilities <https://github.com/ENCODE-DCC/kentUtils.git>`_
    The header files are particularly useful.
"""
import os
import warnings

from plastid.readers.autosql import AutoSqlDeclaration
from plastid.util.io.binary import BinaryParserFactory, find_null_bytes
from plastid.util.services.mini2to3 import safe_bytes, safe_str

#===============================================================================
# Exported warnings
#===============================================================================

WARN_CHROM_NOT_FOUND = "No data for chromosome '%s' in file '%s'."
WARN_FILE_NOT_FOUND = "File '%s' not found."


#===============================================================================
# Functions
#===============================================================================


cdef lm * get_lm(lm * my_lm=NULL, int maxmem=0) except NULL:
    """Allocate or destroy and reallocate a local memory pool, aiming to keep
    it under `maxmem`

    Parameters
    ----------
    my_lm : *lm, optional
        Local memory pool. If NULL, a new pool is allocated. (Default: NULL)

    maxmem : int, optional
        Maximum desired size in bytes for memory pool. If the size of `my_lm`
        exceeds `maxmem`, `my_lm` will be freed and a new pool created.


    Returns
    -------
    lm
        local memory buffer, or NULL if memory cannot be allocated

    Raises
    ------
    MemoryError
        If memory cannot be allocated
    """
    cdef:
        int realmaxmem = maxmem >> 10

    if my_lm == NULL:
        my_lm = lmInit(0)

    elif maxmem > 0 and lmSize(my_lm) >= maxmem:
        lmCleanup(&my_lm)
        my_lm = NULL
        my_lm = lmInit(0)
    if not my_lm:
        raise MemoryError("BigBed/BigWig reader: Could not allocate local memory block.")

    return my_lm


#===============================================================================
# Classes
#===============================================================================


cdef class _BBI_Reader:
    """Abstract base class for `BigWig`_ file readers

    Reads basic file properties.

    Parameters
    ----------
    filename : str
        Name of file to open

    maxmem : float
        Maximum desired memory footprint for C objects, in megabytes.
        May be temporarily exceeded if large queries are requested.
        (Default: 0, No maximum)
    """
    def __cinit__(self, str filename, maxmem=0, *args, **kwargs):
        if not os.path.exists(filename):
            raise IOError(WARN_FILE_NOT_FOUND % filename)

        self.filename      = filename
        self._chromlengths = None
        self._chromids     = None
        self._summary      = None
        self._lm           = NULL
        self._maxmem       = int(round(maxmem * 1024 * 1024))

    def __dealloc__(self):
        """Close `BigBed`_/`BigWig`_ file"""

        # deallocate memory, if allocated
        if self._lm != NULL:
            lmCleanup(&self._lm)

        # close file
        close_file(self._bbifile)

    def __repr__(self):
        return "<%s file='%s'>" % (self.__class__.__name__, self.filename)

    def __str__(self):
        return repr(self)

    cdef lm * _get_lm(self):
        """Return ``self._lm``, allocating it if necessary

        Returns
        -------
        lm
            local memory buffer

        Raises
        ------
        MemoryError
            If memory cannot be allocated
        """
        self._lm = get_lm(my_lm=self._lm, maxmem=self._maxmem)
        return self._lm

    property filename:
        """Name of BigWig or BigBed file"""
        def __get__(self):
            return self.filename

    property version:
        """Version of BigWig or BigBed file format"""
        def __get__(self):
            return self._bbifile.version

    property chroms:
        """Dictionary mapping chromosome names to lengths"""
        def __get__(self):
            return self.c_chroms()

    property chrom_sizes:
        """DEPRECATED: Use `.chroms` instead of `.chrom_sizes`"""
        def __get__(self):
            warnings.warn(
                "`chrom_sizes` is deprecated and will be removed from plastid v0.6.0. Use `chroms` instead",
                UserWarning
            )
            return self.c_chroms()

    # FIXME: hide
    property chromids:
        def __get__(self):
            return self._chromids
# see note below.
#     property summary_info:
#         """Summary information over total BBI file"""
#         def __get__(self):
#             if self._summary is not None:
#                 return self._summary
#             else:
#                 return self.fetch_summary()

    property uncompress_buf_size:
        """Size of buffer needed to uncompress blocks. If 0, the data is uncompressed"""
        def __get__(self):
            return self._bbifile.uncompressBufSize

# data for mean as read from table is very inaccurate. Decide what to do later
#     cdef dict fetch_summary(self):
#         """Return summary info of BBI file, and set `self._summary`.
#         Summary info includes:
#
#           - min value
#           - max value
#           - sum of values
#           - sum of squares of values
#           - covered bases
#           - mean value over covered bases
#
#         Returns
#         -------
#         dict
#             dictionary described above
#         """
#         cdef:
#             bbiSummaryElement mydata = bbiTotalSummary(self._bbifile)
#             dict dtmp
#
#         dtmp = {
#             "min" : mydata.minVal,
#             "max" : mydata.maxVal,
#             "mean" : mydata.sumData / mydata.validCount,
#             "sum"  : mydata.sumData,
#             "sum_squares" : mydata.sumSquares,
#             "valid_bases" : mydata.validCount,
#         }
#         self._summary = dtmp
#         return dtmp

    cdef dict _define_chroms(self):
        """Return dictionary mapping chromosome names to their lengths

        Returns
        -------
        dict
            Dictionary mapping chromosome names to their lengths
        """
        cdef:
            dict cdict = {}
            dict idict = {}
            str chr_name
            int chr_size
            bbiChromInfo *chrom_info = bbiChromList(self._bbifile)

        while chrom_info is not NULL:
            chr_name             = safe_str(chrom_info.name)
            chr_size             = chrom_info.size
            cdict[chr_name]      = chr_size
            idict[chrom_info.id] = chr_name
            chrom_info           = chrom_info.next

        self._chromlengths = cdict
        self._chromids     = idict
        bbiChromInfoFreeList(&chrom_info)

        return cdict

    cdef dict c_chroms(self):
        """Return dictionary mapping chromosome names to their lengths,
        populating ``self._chrominfo`` if necessary

        Returns
        -------
        dict
            Dictionary mapping chromosome names to their lengths
        """
        if self._chromlengths is not None:
            return self._chromlengths
        else:
            self._define_chroms()
            return self._chromlengths




#!/usr/bin/env python
"""This module contains classes that:

  - build :class:`argparse.ArgumentParser` objects for various data types
    used in genomics

  - parse those arguments into useful file types


Arguments are grouped into the following sets:

    ===========================================================   ======================================
    **Parameter/argument set**                                    **Parser building class**
    -----------------------------------------------------------   --------------------------------------
    Generic parameters (e.g. for error reporting, logging)        :class:`BaseParser`

    :term:`Read alignments` or :term:`count files <count file>`   :class:`AlignmentParser`

    Genomic feature or mask annotations                           :class:`AnnotationParser`

    Genomic sequence files                                        :class:`SequenceParser`

    Plotting parameters for charts                                :class:`PlottingParser`
    ===========================================================   ======================================


Example
-------
To use any of these in your own command line scripts, follow these steps:

  #. Import one or more of the classes above::

         >>> import argparse
         >>> from plastid.util.scriptlib.argparsers import AnnotationParser


  #. Use the first function to create an  :class:`~argparse.ArgumentParser`,
     and supply this object as a `parent` when you build your script's
     :py:class:`~argparse.ArgumentParser`::

         >>> ap = AnnotationParser()

         # create annotation file parser
         >>> annotation_file_parser = ap.get_parser(disabled=["some_option_to_disable"])

         # create my own parser, incorporating flags from annotation_file_parser
         >>>> my_own_parser = argparse.ArgumentParser(parents=[annotation_file_parser])

         # add script-specific arguments
         >>> my_own_parer.add_argument("positional_argument",type=str)
         >>> my_own_parser.add_argument("--foo",type=int,default=5,help="Some option")
         >>> my_own_parser.add_argument("--bar",type=str,default="a string",help="Another option")

  #. Then, use the second parse the arguments::

         >>> args = parser.parse_args()

         # get transcript objects from arguments
         # this will be an iterator over |Transcripts|
         >>> transcripts = ap.get_transcripts_from_args(args)

         >>> pass # rest of your script


Your script will then be able process whatever sorts of annotation files that
plastid currently supports.


See Also
--------
:py:mod:`argparse`
    Python documentation on argument parsing

:py:obj:`plastid.bin`
    Source code of command-line scripts, for further examples
"""
import sys
import os
import functools
import argparse
import warnings
import pkg_resources
import pysam

from plastid.util.services.exceptions import (
    MalformedFileError,
    ArgumentWarning,
    DataWarning,
    FileFormatWarning,
    filterwarnings,
)
from plastid.util.services.decorators import deprecated
from plastid.util.io.openers import opener, NullWriter
from plastid.util.io.filters import CommentReader
from plastid.genomics.roitools import Transcript, SegmentChain
from plastid.readers.gff import (
    _DEFAULT_GFF3_TRANSCRIPT_TYPES,
    _DEFAULT_GFF3_EXON_TYPES,
    _DEFAULT_GFF3_CDS_TYPES,
)

#===============================================================================
# INDEX: Constants used in parsers below
#===============================================================================

# Size above which we recommend sorting a GFF/GTF2 file
_GFF_SORT_SIZE = 100 * 1024 * 1024

_MAPPING_RULE_TITLE = "alignment mapping functions (BAM & bowtie files only)"
_MAPPING_RULE_DESCRIPTION = \
"""For BAM or bowtie files, one of the mutually exclusive read mapping functions
is required:
"""

_MAPPING_OPTION_TITLE = "filtering and alignment mapping options"
_MAPPING_OPTION_DESCRIPTION = \
"""
The remaining arguments are optional and affect the behavior of specific
mapping functions:
"""


_DEFAULT_ALIGNMENT_FILE_PARSER_DESCRIPTION = \
"Open alignment or count files and optionally set mapping rules"

_DEFAULT_ALIGNMENT_FILE_PARSER_TITLE = "count & alignment file options"



_DEFAULT_ANNOTATION_PARSER_DESCRIPTION = \
"Open one or more genome annotation files"

_DEFAULT_ANNOTATION_PARSER_TITLE = \
"annotation file options (one or more annotation files required)"

GFF_SORT_MESSAGE = \
"""Sort and index your GTF2/GFF with Tabix as follows:

    $ sort -k1,1 -k4,4n my_file.FORMAT | bgzip > my_file_sorted.FORMAT.gz
    $ tabix -p gff my_file_sorted.FORMAT.gz

See http://www.htslib.org/doc/tabix.html for download and documentation of
tabix and bgzip."""

_MASK_PARSER_TITLE = "mask file options (optional)"
_MASK_PARSER_DESCRIPTION = \
"""Add mask file(s) that annotate regions that should be excluded from analyses
(e.g. repetitive genomic regions)."""

_DEFAULT_SEQUENCE_PARSER_TITLE = "sequence options"
_DEFAULT_SEQUENCE_PARSER_DESCRIPTION = ""

_DEFAULT_PLOTTING_TITLE = "Plotting options"

#===============================================================================
# INDEX: Base class for parsers
#===============================================================================


class Parser(object):
    """Base class for argument parser factories used below

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes
    """

    def __init__(self, groupname=None, prefix="", disabled=None, **kwargs):
        """Create a parser

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes
        """
        self.prefix = prefix
        self.disabled = [] if disabled is None else disabled
        self.groupname = groupname

        # define in __init__ of subclass
        self.arguments = []

    def get_parser(
            self, parser=None, groupname=None, arglist=None, title=None, description=None, **kwargs
    ):
        """Create an populate :class:`argparse.ArgumentParser` with arguments

        Parameters
        ----------
        parser : :class:`argparse.ArgumentParser` or None, optional
            If `None`, a new parser will be created, and arguments will be
            added to it. If not `None`, arguments will be added to `parser`.
            (Default: `None`)

        groupname : str or None, optional
            If not `None`, default to `self.groupname`. If either `groupname`
            or `self.groupname` is not `None`, an option group with this name
            will be added to `parser`, and arguments added to that groupname
            instead of the main argument group of `parser`. In this case,
            `title` and `description` will be applied to the option group
            instead of to `parser`.  Default : `None`)

        arglist : list, optional
            If not `None`, arguments in this list will be added to `parser`.
            Otherwise, arguments will be taken from `self.arguments`.

            The list should be a list of tuples of
            ('argument_name',dict_of_options), where `argument_name` is a
            string, and `dict_of_options` a dictionary of keyword arguments to
            pass to :meth:`argparse.ArgumentParser.add_argument`.

        title : str, optional
            Optional title for parser

        description : str, optional
            Optional description for parser


        kwargs : keyword arguments
            Additional arguments passed during creation of
            :class:`argparse.ArgumentParser`


        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        if groupname is None:
            groupname = self.groupname

        if parser is None:
            if groupname is None:
                parser = argparse.ArgumentParser(description=description, add_help=False, **kwargs)
            else:
                parser = argparse.ArgumentParser(add_help=False, **kwargs)

        addto = parser
        if groupname is not None:
            addto = parser.add_argument_group(title=title, description=description)

        arglist = self.arguments if arglist is None else arglist
        for arg_name, arg_opts in filter(lambda x: x[0] not in self.disabled, arglist):
            addto.add_argument("--%s%s" % (self.prefix, arg_name), **arg_opts)

        return parser


#===============================================================================
# INDEX: Alignment & count file parser
#===============================================================================


class AlignmentParser(Parser):
    """Parser for files containing read alignments or quantitative data.

    Checks for additional mapping rules and command-line arguments
    by checking the entrypoints ``plastid.mapping_rules`` and
    ``plastid.mapping_options``


    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes

    input_choices : list, optional
        list of permitted alignment file type choices for input

    allow_mapping : bool, optional
        Enable/disable user configuration of mapping rules (default: True)
    """
    def __init__(
            self,
            prefix="",
            disabled=None,
            input_choices=("BAM", "bigwig", "bowtie", "wiggle"),
            groupname="alignment_options",
            allow_mapping=True
    ): # yapf: disable
        """Create a parser for read alignments and/or quantitative data

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes

        input_choices : list, optional
            list of permitted alignment file type choices for input

        allow_mapping : bool, optional
            Enable/disable user configuration of mapping rules (default: True)
        """
        Parser.__init__(self, groupname=groupname, prefix=prefix, disabled=disabled)
        self.input_choices = input_choices
        self.allow_mapping = allow_mapping

        self.bamfuncs = {}
        self.bowtiefuncs = {}

        self.arguments = [
            ("count_files"     , dict(type=str,
                                      default=[],
                                      nargs="+",
                                      help="One or more count or alignment file(s) from a single sample or set of samples to be pooled.")),
            ("countfile_format", dict(choices=input_choices,
                                      default="BAM",
                                      help="Format of file containing alignments or counts (Default: %(default)s)")),
            ("normalize"        , dict(action="store_true",
                                       help="Whether counts should be normalized"+
                                            " to counts per million (usually not. default: %(default)s)",
                                       default=False)),
            ("sum"              , dict(type=float,default=None,
                                       help="Sum used in normalization of counts and RPKM/RPNT calculations "+\
                                            "(Default: total mapped reads/counts in dataset)")),

            ]

        length_ops = [
            (
                "min_length",
                dict(
                    type=int,
                    default=25,
                    metavar="N",
                    help=(
                        "Minimum read length required to be included "
                        "(BAM & bowtie files only. Default: %(default)s)"
                    ),
                ),
            ),
            (
                "max_length",
                dict(
                    type=int,
                    default=100,
                    metavar="N",
                    help=(
                        "Maximum read length permitted to be included "
                        "(BAM & bowtie files only. Default: %(default)s)"
                    ),
                ),
            ),
        ]

        big_genome = [
            (
                "big_genome",
                dict(
                    action="store_true",
                    default=False,
                    help=(
                        "Use slower but memory-efficient implementation "
                        "for big genomes or for memory-limited computers. "
                        "For wiggle & bowtie files only."
                    ),
                ),
            ),
        ]

        maxmem = [
            (
                "maxmem",
                dict(
                    type=float,
                    default=0,
                    help=(
                        "Maximum desired memory footprint in MB to devote to "
                        "BigBed/BigWig files. May be exceeded by large queries. "
                        "(Default: 0, No maximum)"
                    )
                )
            ),
        ]

        # filetype-specific options
        self.filetype_options = {
            "BAM": length_ops,
            "bowtie": length_ops + big_genome,
            "wiggle": big_genome,
            "bigwig": maxmem,
        }

        if self.allow_mapping == False:
            self.map_rules = []
            self.map_ops = []
        else:
            map_rules = [
                (
                    "fiveprime_variable",
                    dict(
                        action="store_const",
                        const="fiveprime_variable",
                        dest="%smapping" % prefix,
                        help=(
                            "Map read alignment to a variable offset from 5' "
                            "position of read, with offset determined by read "
                            "length. Requires `--offset` below"
                        ),
                    ),
                ),
                (
                    "fiveprime",
                    dict(
                        action="store_const",
                        const="fiveprime",
                        dest="%smapping" % prefix,
                        help="Map read alignment to 5' position."
                    ),
                ),
                (
                    "threeprime",
                    dict(
                        action="store_const",
                        const="threeprime",
                        dest="%smapping" % prefix,
                        help="Map read alignment to 3' position"
                    ),
                ),
                (
                    "center",
                    dict(
                        action="store_const",
                        const="center",
                        dest="%smapping" % prefix,
                        help=(
                            "Subtract N positions from each end of read, "
                            "and add 1/(length-N), to each remaining position, "
                            "where N is specified by `--nibble`"
                        )
                    )
                ),
            ]
            map_ops = [
                (
                    "offset",
                    dict(
                        default=0,
                        metavar="OFFSET",
                        help=( 
                            "For `--fiveprime` or `--threeprime`, provide an integer "
                            "representing the offset into the read, starting "
                            "from either the 5\' or 3\' end, at which data "
                            "should be plotted. For `--fiveprime_variable`, "
                            "provide the filename of a two-column tab-delimited "
                            "text file, in which first column represents read "
                            "length or the special keyword `'default'`, and "
                            "the second column represents the offset from the "
                            "five prime end of that read length at which the "
                            "read should be mapped. (Default: %(default)s)"
                        ),
                    ),
                ),
                (
                    "nibble",
                    dict(
                        type=int,
                        default=0,
                        metavar="N",
                        help=(
                            "For use with `--center` only. nt to remove from "
                            "each end of read before mapping "
                            "(Default: %(default)s)"
                        ),
                    ),
                ),
            ]

            for epoint in pkg_resources.iter_entry_points(group="plastid.mapping_rules"):
                reg_name = epoint.name
                pdict = epoint.load()
                if "name" in pdict:
                    reg_name = pdict.pop("name")

                pdict["const"] = reg_name
                pdict["action"] = "store_const"
                pdict["dest"] = "%smapping" % self.prefix

                bamfunc = pdict.get("bamfunc", None)
                bowtiefunc = pdict.get("bowtiefunc", None)

                if bamfunc is not None:
                    self.bamfuncs[reg_name] = bamfunc
                    pdict.pop("bamfunc")

                if bowtiefunc is not None:
                    self.bowtiefuncs[reg_name] = bowtiefunc
                    pdict.pop("bowtiefunc")

                map_rules.append((reg_name, pdict))

            for epoint in pkg_resources.iter_entry_points(group="plastid.mapping_options"):
                reg_name = epoint.name
                pdict = epoint.load()
                if "name" in pdict:
                    reg_name = pdict.pop("name")

                map_ops.append((reg_name, pdict))

            self.map_rules = map_rules
            self.map_ops = map_ops

    def get_parser(
            self,
            title=_DEFAULT_ALIGNMENT_FILE_PARSER_TITLE,
            description=_DEFAULT_ALIGNMENT_FILE_PARSER_DESCRIPTION,
            **kwargs
    ):
        """Return an :py:class:`~argparse.ArgumentParser` that opens alignment
        (`BAM`_,  or `bowtie`_) or count (`Wiggle`_, `bedGraph`_) files.

        In the case of `bowtie`_ or `BAM`_ import, also parse arguments for
        mapping rules (e.g. fiveprime end mapping, threeprime end mapping, et
        c) and optional read length filters


        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)

        kwargs : keyword arguments
            Additional arguments to pass to :meth:`Parser.get_parser`


        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        parser = Parser.get_parser(self, title=title, description=description, **kwargs)
        extra_args = []
        for k in self.input_choices:
            arglist = self.filetype_options.get(k, [])
            for arg in arglist:
                if arg not in extra_args:
                    extra_args.append(arg)

        if len(extra_args) > 0:
            # use mutator function- add new parser to `parser`
            Parser.get_parser(
                self,
                parser=parser,
                arglist=extra_args,
                title=title,
                description=description,
            )

        if self.allow_mapping == True:
            Parser.get_parser(
                self,
                parser=parser,
                groupname="mapping_options",
                arglist=self.map_rules,
                title=_MAPPING_RULE_TITLE,
                description=_MAPPING_RULE_DESCRIPTION,
            )
            Parser.get_parser(
                self,
                parser=parser,
                groupname="sub_options",
                arglist=self.map_ops,
                title=_MAPPING_OPTION_TITLE,
                description=_MAPPING_OPTION_DESCRIPTION,
            )

        return parser

    def get_genome_array_from_args(self, args, printer=None):
        """Return a |GenomeArray|, |SparseGenomeArray| or |BAMGenomeArray|
        from arguments parsed by :py:func:`get_alignment_file_parser`

        Parameters
        ----------
        args : :py:class:`argparse.Namespace`
            Arguments from the parser

        printer : file-like, optional
            A stream to which stderr-like info can be written (default:
            |NullWriter|)


        Returns
        -------
        |GenomeArray|, |SparseGenomeArray|, or |BAMGenomeArray|
        """
        from plastid.genomics.genome_array import (
            GenomeArray,
            SparseGenomeArray,
            BAMGenomeArray,
            BigWigGenomeArray,
            SizeFilterFactory,
            CenterMapFactory,
            FivePrimeMapFactory,
            ThreePrimeMapFactory,
            VariableFivePrimeMapFactory,
            five_prime_map,
            three_prime_map,
            center_map,
            variable_five_prime_map,
        )

        args = PrefixNamespaceWrapper(args, self.prefix)
        disabled = self.disabled
        map_rule = args.mapping

        if printer is None:
            printer = NullWriter()

        # require at least one countfile
        if len(args.count_files) == 0:
            printer.write("Please include at least one input file.")
            sys.exit(1)

        # require mapping rules unless wiggle
        if map_rule is None and args.countfile_format in ("BAM", "bowtie"):
            printer.write("Please specify a read mapping rule.")
            sys.exit(1)

        if "countfile_format" not in disabled:

            if args.countfile_format in ("BAM", "CRAM"):
                count_files = [pysam.Samfile(X, "rb") for X in args.count_files]
                try:
                    ga = BAMGenomeArray(count_files)
                except ValueError:
                    printer.write("Input BAM file(s) not indexed. Please index via:")
                    printer.write("")
                    for fn in args.count_files:
                        printer.write("    samtools index [-b|-c] %s" % fn)
                    printer.write("")
                    printer.write("Exiting.")
                    sys.exit(1)

                size_filter = SizeFilterFactory(min=args.min_length, max=args.max_length)
                ga.add_filter("size:%s-%s" % (args.min_length, args.max_length), size_filter)
                if map_rule == "fiveprime":
                    map_function = FivePrimeMapFactory(int(args.offset))
                elif map_rule == "threeprime":
                    map_function = ThreePrimeMapFactory(int(args.offset))
                elif map_rule == "center":
                    map_function = CenterMapFactory(args.nibble)
                elif map_rule == "fiveprime_variable":
                    if str(args.offset) == "0":
                        printer.write(
                            "Please specify a filename to use for fiveprime "
                            "variable offsets in --offset."
                        )
                        sys.exit(1)
                    offset_dict = _parse_variable_offset_file(CommentReader(open(args.offset)))
                    map_function = VariableFivePrimeMapFactory(offset_dict)
                elif map_rule in self.bamfuncs:
                    map_function = functools.partial(self.bamfuncs[map_rule], args=args)
                else:
                    printer.write(
                        "Mapping rule '%s' not implemented for BAM input. "
                        "Exiting." % map_rule
                    )
                    sys.exit(1)
                ga.set_mapping(map_function)

            elif args.countfile_format == "bigwig":
                ga = BigWigGenomeArray(maxmem=args.maxmem)
                for align_file in args.count_files:
                    ga.add_from_bigwig("%s_fw.bw" % align_file, "+")
                    ga.add_from_bigwig("%s_rc.bw" % align_file, "-")

            # wiggle/bedGraph and bowtie
            else:
                if "big_genome" not in disabled and args.big_genome == True:
                    ga = SparseGenomeArray()
                else:
                    ga = GenomeArray()

                # wiggle/bedGraph
                if args.countfile_format == "wiggle":
                    for align_file in args.count_files:
                        printer.write("Opening wiggle files %s ..." % align_file)
                        with open("%s_fw.wig" % align_file) as fh:
                            ga.add_from_wiggle(fh, "+")
                        with open("%s_rc.wig" % align_file) as fh:
                            ga.add_from_wiggle(fh, "-")

                # bowtie
                elif args.countfile_format == "bowtie":
                    trans_args = {"nibble": int(args.nibble)}
                    if map_rule == "fiveprime_variable":
                        transformation = variable_five_prime_map
                        if str(args.offset) == "0":
                            printer.write(
                                "Please specify a filename to use for "
                                "fiveprime variable offsets in --offset."
                            )
                            sys.exit(1)
                        else:
                            with open(args.offset) as myfile:
                                trans_args["offset"] = _parse_variable_offset_file(
                                    CommentReader(myfile)
                                )
                    else:
                        trans_args["offset"] = int(args.offset)
                        if map_rule == "fiveprime":
                            transformation = five_prime_map
                        elif map_rule == "threeprime":
                            transformation = three_prime_map
                        elif map_rule == "entire":
                            transformation = center_map
                        elif map_rule == "center":
                            transformation = center_map
                        elif map_rule in self.bowtiefuncs:
                            transformation = self.bowtiefuncs[map_rule]
                            trans_args["args"] = args
                        else:
                            printer.write(
                                "Mapping rule '%s' not implemented for bowtie "
                                "input. Exiting." % map_rule
                            )
                            sys.exit(1)

                    for infile in args.count_files:
                        with opener(infile) as my_file:
                            ga.add_from_bowtie(
                                my_file,
                                transformation,
                                min_length=args.min_length,
                                max_length=args.max_length,
                                **trans_args
                            )

        printer.write("Counted %s total reads." % ga.sum())

        if "sum" not in disabled and args.sum is not None:
            ga.set_sum(args.sum)

        if "normalize" not in disabled and args.normalize == True:
            printer.write("Normalizing to reads per million.")
            ga.set_normalize(True)

        return ga


#===============================================================================
# INDEX: Annotation file parser
#===============================================================================


class AnnotationParser(Parser):
    """Parser for annotation files in various formats

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes

    input_choices : list, optional
        list of permitted alignment file type choices for input

    allow_mapping : bool, optional
        Enable/disable user configuration of mapping rules (default: True)
    """

    def __init__(
            self,
            prefix="",
            disabled=None,
            groupname="annotation_options",
            input_choices=("BED", "BigBed", "GTF2", "GFF3")
    ):
        """Create a parser for genomic features in an annotation file

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes

        input_choices : list, optional
            list of permitted alignment file type choices for input

        allow_mapping : bool, optional
            Enable/disable user configuration of mapping rules (default: True)
        """
        Parser.__init__(self, groupname=groupname, prefix=prefix, disabled=disabled)
        self.input_choices = input_choices
        self.arguments = [
            (
                "annotation_files",
                dict(
                    metavar="infile.[%s]" % " | ".join(input_choices),  # | psl]",
                    type=str,
                    nargs="+",
                    default=[],
                    help="Zero or more annotation files (max 1 file if BigBed)"
                )
            ),
            (
                "annotation_format",
                dict(
                    choices=input_choices,
                    default="GTF2",
                    help=(
                        "Format of %sannotation_files (Default: GTF2). "
                        "Note: GFF3 assembly assumes SO v.2.5.2 feature "
                        "ontologies, which may or may not match your specific "
                        "file." % prefix
                    )
                )
            ),
            (
                "add_three",
                dict(
                    default=False,
                    action="store_true",
                    help=(
                        "If supplied, coding regions will be extended by 3 "
                        "nucleotides at their 3\' ends (except for GTF2 files "
                        "that explicitly include `stop_codon` features). "
                        "Use if your annotation file excludes stop codons "
                        "from CDS."
                    ),
                )
            ),
            (
                "tabix",
                dict(
                    default=False,
                    action="store_true",
                    help=(
                        "%sannotation_files are tabix-compressed and indexed "
                        "(Default: False). Ignored for BigBed files."
                        % prefix
                    )
                )
            ),
            (
                "sorted",
                dict(
                    default=False,
                    action="store_true",
                    help=(
                        "%sannotation_files are sorted by chromosomal position "
                        "(Default: False)" % prefix
                    )
                )
            ),
        ]

        # options for specific filetypes
        self.filetype_options = {
            "BED" : [("bed_extra_columns", dict(default=0,
                                                nargs="+",
                                                help="Number of extra columns in BED file (e.g. in custom ENCODE formats) "+
                                                     "or list of names for those columns. (Default: %(default)s)."))
                    ],
            "BigBed" : [
                ("maxmem"           , dict(type=float,default=0,
                                           help="Maximum desired memory footprint in MB to devote to BigBed/BigWig files. May be exceeded by large queries. (Default: 0, No maximum)")),

                        ],
            "GFF3" : [("gff_transcript_types", dict(type=str,
                                                    default=_DEFAULT_GFF3_TRANSCRIPT_TYPES,
                                                    nargs="+",
                                                    help="GFF3 feature types to include as transcripts, even "+\
                                                         "if no exons are present (for GFF3 only; default: use SO v2.5.3 specification)")),
                      ("gff_exon_types", dict(type=str,
                                              default=_DEFAULT_GFF3_EXON_TYPES,
                                              nargs="+",
                                              help="GFF3 feature types to include as exons (for GFF3 only; default: use SO v2.5.3 specification)")),
                      ("gff_cds_types", dict(type=str,
                                             default=_DEFAULT_GFF3_CDS_TYPES,
                                             nargs="+",
                                             help="GFF3 feature types to include as CDS (for GFF3 only; default: use SO v2.5.3 specification)")),
                     ]
             }

    def get_parser(
            self,
            title=_DEFAULT_ANNOTATION_PARSER_TITLE,
            description=_DEFAULT_ANNOTATION_PARSER_DESCRIPTION,
            **kwargs
    ):
        """Return an :class:`~argparse.ArgumentParser` that opens annotation files.

        Parameters
        ----------
        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)

        kwargs : keyword arguments
            Additional arguments to pass to :meth:`Parser.get_parser`

        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        parser = Parser.get_parser(self, title=title, description=description, **kwargs)

        for k in self.input_choices:
            arglist = self.filetype_options.get(k)
            if arglist is not None:
                # use mutator function- add new parser to `parser`
                Parser.get_parser(
                    self,
                    parser=parser,
                    groupname="%s_%s_options" % (self.groupname, k),
                    title="%s-specific options" % k,
                    arglist=arglist
                )

        return parser

    def get_transcripts_from_args(self, args, printer=None, return_type=None, require_sort=False):
        """Return a generator of |Transcript| objects from arguments parsed by
        :func:`get_annotation_file_parser`

        Parameters
        ----------
        args : :py:class:`argparse.Namespace`
            Namespace object from :py:func:`get_annotation_file_parser`

        printer : file-like, optional
            A stream to which stderr-like info can be written (Default: |NullWriter|)

        return_type : |SegmentChain| or subclass, optional
            Type of object to return (Default: |Transcript|)

        require_sort : bool, optional
            If True, quit if the annotation file(s) are not sorted or indexed

        Returns
        -------
        iterator
            |Transcript| objects, either in order of appearance (if input was a
            `BED`_, `BigBed`_, or `PSL`_ file), or sorted lexically by
            chromosome, start coordinate, end coordinate, and then strand (if
            input was `GTF2`_ or `GFF3`_).


        See Also
        --------
        get_annotation_file_parser
            Function that creates :py:class:`argparse.ArgumentParser` whose
            output :py:class:`~argparse.Namespace` is processed by this
            function
        """

        return self.get_segmentchains_from_args(
            args, printer=printer, return_type=Transcript, require_sort=require_sort
        )

    def get_segmentchains_from_args(self, args, printer=None, return_type=None, require_sort=False):
        """Return a generator of |SegmentChain| objects from arguments parsed
        by :py:func:`get_annotation_file_parser`

        Parameters
        ----------
        args : :py:class:`argparse.Namespace`
            Namespace object from :py:func:`get_annotation_file_parser`

        printer : file-like, optional
            A stream to which stderr-like info can be written (Default: |NullWriter|)

        return_type : |SegmentChain| or subclass, optional
            Type of object to return (Default: |Transcript|)

        require_sort : bool, optional
            If True, quit if the annotation file(s) are not sorted or indexed

        Returns
        -------
        iterator
            |SegmentChain| objects, either in order of appearance (if input was a
            `BED`_, `BigBed`_, or `PSL`_ file), or sorted lexically by chromosome,
            start coordinate, end coordinate, and then strand (if input was `GTF2`_
            or `GFF3`_).


        See Also
        --------
        get_annotation_file_parser
            Function that creates :py:class:`argparse.ArgumentParser` whose output
            :py:class:`~argparse.Namespace` is processed by this function
        """
        if printer is None:
            printer = NullWriter()

        if return_type is None:
            return_type = SegmentChain

        args = PrefixNamespaceWrapper(args, self.prefix)
        disabled = self.disabled

        if require_sort == True and 'sorted' not in disabled:
            if args.annotation_format in ("BED", "GTF2", "GFF3") \
                and args.sorted == False \
                and 'tabix' not in disabled \
                and args.tabix == False:

                printer.write(
                    "Using unsorted/unindexed annotation files requires "
                    "impractical amounts of memory."
                )
                if args.annotation_format == "BED":
                    printer.write(
                        """Convert BED to BigBed using Jim Kent's bedToBigBed utility as follows:

        $ sort -k1,1 -k2,2n my_file > my_file_sorted.bed
        $ bedToBigBed my_file_sorted.bed chrom.sizes my_file_sorted.bb

    See https://github.com/ENCODE-DCC/kentUtils/tree/master/src/product/scripts
    for download & documentation of Kent utilities"""
                    )
                    sys.exit(1)
                else:
                    printer.write(GFF_SORT_MESSAGE.replace("FORMAT", args.annotation_format))
                    sys.exit(1)

        printer.write("Parsing features in %s ..." % ", ".join(args.annotation_files))

        if "tabix" not in disabled:
            tabix = args.tabix
        else:
            tabix = False

        if "add_three" not in disabled:
            add_three = args.add_three
        else:
            add_three = False

        if "bed_extra_columns" not in disabled:
            bed_extra_columns = args.bed_extra_columns
            if not (isinstance(bed_extra_columns, list)):
                try:
                    bed_extra_columns = int(bed_extra_columns)
                except ValueError:
                    pass
        else:
            bed_extra_columns = 0

        if args.annotation_format.lower() == "bigbed":
            if len(args.annotation_files) > 1:
                printer.write("Bad arguments: we can only process one BigBed file.")
                sys.exit(2)
            if tabix == True:
                warnings.warn(
                    "Tabix compression is incompatible with BigBed files. Ignoring.",
                    ArgumentWarning,
                )

            from plastid.readers.bigbed import BigBedReader
            transcripts = iter(
                BigBedReader(
                    args.annotation_files[0],
                    return_type=return_type,
                    cache_depth=1,
                    add_three_for_stop=add_three,
                    printer=printer,
                    maxmem=args.maxmem
                )
            )

        elif tabix == True:
            streams = [
                pysam.tabix_iterator(opener(X), pysam.asTuple()) for X in args.annotation_files
            ]
        else:
            streams = (opener(X) for X in args.annotation_files)

        if args.annotation_format in ("GFF3", "GTF2"):
            from plastid.readers.gff import GFF3_TranscriptAssembler, GTF2_TranscriptAssembler
            if 'sorted' not in disabled \
                and args.sorted == False \
                and 'tabix' not in disabled \
                and args.tabix == False \
                and any((os.stat(X).st_size >= _GFF_SORT_SIZE for X in args.annotation_files)):

                msg = (
                    "Transcript assembly on large FORMAT files can require "
                    "a lot of memory. Consider using a sorted file with "
                    "the '--sorted' flag and/or tabix-compression. "
                )
                msg += GFF_SORT_MESSAGE
                msg = msg.replace("FORMAT", args.annotation_format)
                warnings.warn(msg, ArgumentWarning)

        if args.annotation_format.lower() == "gff3":
            transcripts = GFF3_TranscriptAssembler(
                *streams,
                transcript_types=args.gff_transcript_types,
                exon_types=args.gff_exon_types,
                cds_types=args.gff_cds_types,
                printer=printer,
                add_three_for_stop=add_three,
                tabix=tabix,
                return_type=return_type,
                is_sorted=args.sorted
            )
        elif args.annotation_format.lower() == "gtf2":
            transcripts = GTF2_TranscriptAssembler(
                *streams,
                printer=printer,
                tabix=tabix,
                return_type=return_type,
                add_three_for_stop=add_three,
                is_sorted=args.sorted
            )

        elif args.annotation_format.lower() == "bed":
            from plastid.readers.bed import BED_Reader
            transcripts = BED_Reader(
                *streams,
                add_three_for_stop=add_three,
                tabix=tabix,
                return_type=return_type,
                printer=printer,
                extra_columns=bed_extra_columns
            )

        elif args.annotation_format.lower() == "psl":
            from plastid.readers.psl import PSL_Reader
            transcripts = PSL_Reader(
                *streams, tabix=tabix, return_type=return_type, printer=printer
            )

        return transcripts

    def get_genome_hash_from_args(self, args, printer=None):
        """Return a |GenomeHash| of regions from command-line arguments

        Parameters
        ----------
        args : :py:class:`argparse.Namespace`
            Namespace object from :py:func:`get_mask_file_parser`

        printer : file-like
            A stream to which stderr-like info can be written (Default: |NullWriter|)


        Returns
        -------
        |GenomeHash|
            Hashed data structure of masked genomic regions


        See Also
        --------
        get_mask_file_parser
            Function that creates :py:class:`argparse.ArgumentParser` whose output
            :py:class:`~argparse.Namespace` is processed by this function
        """
        from plastid.genomics.genome_hash import GenomeHash, BigBedGenomeHash, TabixGenomeHash
        from plastid.readers.bed import BED_Reader
        from plastid.readers.gff import GTF2_Reader, GFF3_Reader
        from plastid.readers.psl import PSL_Reader
        if printer is None:
            printer = NullWriter()

        prefix = self.prefix
        args = PrefixNamespaceWrapper(args, prefix)

        if len(args.annotation_files) > 0:
            printer.write(
                "Opening mask annotation file(s) %s ..." % ", ".join(args.annotation_files)
            )
            if args.annotation_format in ("BED", "GTF2", "GFF3") and args.tabix == False:
                msg = (
                    "Unindexed mask files can require lots of memory in large "
                    "(e.g. mammalian) genomes. Consider converting to BigBed "
                    "or using tabix to index your mask file."
                )
                warnings.warn(msg, ArgumentWarning)

            if len(args.annotation_files) > 0:
                if args.annotation_format == "BigBed":
                    if len(args.annotation_files) > 1:
                        printer.write("Bad arguments: we can only process one BigBed file.")
                        sys.exit(2)
                    return BigBedGenomeHash(args.annotation_files[0])
                elif "tabix" not in self.disabled and args.tabix == True:
                    return TabixGenomeHash(
                        args.annotation_files, args.annotation_format, printer=printer
                    )
                else:
                    streams = (opener(X) for X in args.annotation_files)
                    if args.annotation_format == "BED":
                        reader = BED_Reader
                    elif args.annotation_format == "GTF2":
                        reader = GTF2_Reader
                    elif args.annotation_format == "GFF3":
                        reader = GFF3_Reader
                    elif args.annotation_format == "PSL":
                        reader = PSL_Reader
                    else:
                        assert False

                    hash_ivcs = list(reader(*streams))

                    return GenomeHash(hash_ivcs)
        else:
            return GenomeHash()


class MaskParser(AnnotationParser):
    """Create a parser for masking genomic features given in an annotation file

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes

    input_choices : list, optional
        list of permitted alignment file type choices for input

    allow_mapping : bool, optional
        Enable/disable user configuration of mapping rules (default: True)
    """

    def __init__(
            self,
            prefix="mask_",
            disabled=None,
            groupname="mask_options",
            input_choices=("BED", "BigBed", "GTF2", "GFF3", "PSL")
    ):
        """Create a parser for genomic features in an annotation file

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes

        input_choices : list, optional
            list of permitted alignment file type choices for input

        allow_mapping : bool, optional
            Enable/disable user configuration of mapping rules (default: True)
        """
        AnnotationParser.__init__(
            self,
            prefix=prefix,
            disabled=disabled,
            groupname=groupname,
            input_choices=("BED", "BigBed", "GTF2", "GFF3", "PSL")
        )

    def get_parser(self, title=_MASK_PARSER_TITLE, description=_MASK_PARSER_DESCRIPTION, **kwargs):
        """Return an :py:class:`~argparse.ArgumentParser` that opens annotation
        files as masks alignment (`BAM`_ or `bowtie`_) or count (`Wiggle`_,
        `bedGraph`_) files.

        Parameters
        ----------
        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)

        arglist : list, optional
            If not `None`, arguments in this list will be added to `parser`.
            Otherwise, arguments will be taken from `self.arguments`.

            The list should be a list of tuples of
            ('argument_name',dict_of_options), where `argument_name` is a
            string, and `dict_of_options` a dictionary of keyword arguments to
            pass to :meth:`argparse.ArgumentParser.add_argument`.

        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        return AnnotationParser.get_parser(self, title=title, description=description, **kwargs)


#===============================================================================
# INDEX: Sequence parser
#===============================================================================


class SequenceParser(AnnotationParser):
    """Parser for sequence files

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes

    input_choices : list, optional
        list of permitted alignment file type choices for input
    """

    def __init__(
            self,
            groupname="sequence_options",
            prefix="",
            disabled=None,
            input_choices=("fasta", "fastq", "twobit", "genbank", "embl"),
    ):
        """Create a parser for genomic sequence

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes

        input_choices : list, optional
            list of permitted alignment file type choices for input
        """
        Parser.__init__(self, groupname=groupname, prefix=prefix, disabled=disabled)
        self.input_choices = input_choices
        self.arguments = [
            (
                "sequence_file",
                dict(
                    metavar="infile.[%s]" % " | ".join(input_choices),
                    type=str,
                    help="A file of DNA sequence"
                )
            ),
            (
                "sequence_format",
                dict(
                    choices=input_choices,
                    default="fasta",
                    help="Format of %ssequence_file (Default: fasta)." % prefix
                )
            ),
        ]

    def get_parser(
            self,
            title=_DEFAULT_SEQUENCE_PARSER_TITLE,
            description=_DEFAULT_SEQUENCE_PARSER_DESCRIPTION,
            **kwargs
    ):
        """Return an :py:class:`~argparse.ArgumentParser` that opens sequence files

        Parameters
        ----------

        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)

        kwargs : keyword arguments
            Additional arguments to pass to :meth:`Parser.get_parser`


        Returns
        -------
        :class:`argparse.ArgumentParser`


        See also
        --------
        get_seqdict_from_args
            function that parses the :py:class:`~argparse.Namespace` returned
            by this :py:class:`~argparse.ArgumentParser`
        """

        return Parser.get_parser(self, title=title, description=description, **kwargs)

    def get_seqdict_from_args(self, args, index=True, printer=None):
        """Retrieve a dictionary-like object of sequences

        Parameters
        ----------
        args : :py:class:`argparse.Namespace`
            Namespace object from :py:func:`get_sequence_file_parser`

        index : bool, optional
            If sequence format is anything other than twobit, open with
            lazily-evaluating :func:`Bio.SeqIO.index` instead of
            :func:`Bio.SeqIO.to_dict` (Default: `True`)

        printer : file-like
            A stream to which stderr-like info can be written (Default: |NullWriter|)

        Returns
        -------
        dict-like
            Dictionary-like object mapping chromosome names to
            :class:`Bio.SeqRecord.SeqRecord`-like objects
        """
        if printer is None:
            printer = NullWriter()

        args = PrefixNamespaceWrapper(args, self.prefix)
        printer.write("Opening sequence file '%s'." % args.sequence_file)
        if args.sequence_format == "twobit":
            from plastid.genomics.seqtools import TwoBitSeqRecordAdaptor
            return TwoBitSeqRecordAdaptor(args.sequence_file)
        else:
            from Bio import SeqIO
            if index == True:
                return SeqIO.index(args.sequence_file, args.sequence_format)
            else:
                return SeqIO.to_dict(SeqIO.parse(args.sequence_file, args.sequence_format))


#===============================================================================
# INDEX: Plotting parser
#===============================================================================


class PlottingParser(Parser):
    """Parser for plotting options

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes
    """

    def __init__(self, groupname="plotting_options", prefix="", disabled=None):
        """Create a parser for plotting arguments

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes
        """
        Parser.__init__(self, groupname=groupname, prefix=prefix, disabled=disabled)
        from matplotlib.backend_bases import FigureCanvasBase as fcb
        if len(prefix) > 0:
            prefix += "_"

        try:
            filetypes = sorted(fcb.get_supported_filetypes().keys())
            default_ftype = fcb.get_default_filetype()
        except:  # matplotlib < 1.4.0
            filetypes = ["eps", "jpeg", "pdf", "png", "svg"]
            default_ftype = "pdf"

        self.arguments = [
            (
                "figformat",
                dict(
                    default=default_ftype,
                    type=str,
                    choices=filetypes,
                    help="File format for figure(s); Default: %(default)s)"
                ),
            ),
            (
                "figsize",
                dict(
                    nargs=2,
                    default=None,
                    type=float,
                    metavar="N",
                    help=(
                        "Figure width and height, in inches. "
                        "(Default: use matplotlibrc params)"
                    ),
                ),
            ),
            ("title", dict(type=str, default=None, help="Base title for plot(s).")),
            (
                "cmap",
                dict(
                    type=str,
                    default=None,
                    help=(
                        "Matplotlib color map from which palette will be made "
                        "(e.g. 'Blues','autumn','Set1'; default: use colors "
                        "from ``--stylesheet`` "
                        "if given, or color cycle in matplotlibrc)"
                    ),
                ),
            ),
            ("dpi", dict(type=int, default=150, help="Figure resolution (Default: %(default)s)")),
        ]

        try:
            import matplotlib.style
            stylesheets = matplotlib.style.available

            if "stylesheet" not in self.disabled:
                self.arguments.append(
                    (
                        "stylesheet",
                        dict(
                            default=None,
                            choices=stylesheets,
                            help=(
                                "Use this matplotlib stylesheet instead "
                                "of matplotlibrc params"
                            ),
                        )
                    )
                )
        except ImportError:  # matplotlib < 1.4.0
            pass

    def get_parser(self, title=_DEFAULT_PLOTTING_TITLE, description=None):
        """Return an :py:class:`~argparse.ArgumentParser` to control plotting

        Parameters
        ----------

        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)


        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        return Parser.get_parser(self, title=title, description=description)

    def set_style_from_args(self, args):
        """Parse style information, if present on system and defined in `args`

        Parameters
        ----------
        args : :class:`argparse.Namespace`
            Namespace object from :func:`get_plotting_parser`
        """
        try:
            import matplotlib.style
            if getattr(args, "stylesheet", None) is not None:
                matplotlib.style.use(args.stylesheet)
        except ImportError:
            pass

    def get_figure_from_args(self, args, **kwargs):
        """Return a :class:`matplotlib.figure.Figure` following arguments from
        :func:`get_plotting_parser`

        A new figure is created with parameters specified in `args`. If these are
        not found, values found in `**kwargs` will instead be used. If these are
        not found, we fall back to matplotlibrc values.

        Parameters
        ----------
        args : :class:`argparse.Namespace`
            Namespace object from :func:`get_plotting_parser`

        kwargs : keyword arguments
            Fallback arguments for items not defined in `args`, plus any other
            keyword arguments.

        Returns
        -------
        :class:`matplotlib.figure.Figure`
            Matplotlib figure
        """
        import matplotlib.pyplot as plt
        args = PrefixNamespaceWrapper(args, self.prefix)

        fargs = {}
        # keep this loop in place in case we add additional
        # command line attributes as fig properties later
        for attr in ("figsize", ):
            if attr in kwargs:
                v = kwargs[attr]
            else:
                v = getattr(args, attr, None)

            if v is not None:
                fargs[attr] = v

        # copy values from fargs
        kwargs.update(fargs)
        return plt.figure(**kwargs)

    def get_colors_from_args(self, args, num_colors):
        """Return a list of colors from arguments parsed by a parser from
        :func:`get_plotting_parser`

        If a matplotlib colormap is specified in `args.figcolors`, colors will be
        generated from that map.

        Otherwise, if a stylesheet is specified, colors will be fetched from
        the stylesheet's color cycle.

        Otherwise, colors will be chosen from the default color cycle specified
        ``matplotlibrc``.


        Parameters
        ----------
        args : :class:`argparse.Namespace`
            Namespace object from :func:`get_plotting_parser`

        num_colors : int
            Number of colors to fetch


        Returns
        -------
        list
            List of matplotlib colors
        """
        import matplotlib.cm
        args = PrefixNamespaceWrapper(args, self.prefix)
        figcolors = getattr(args, "cmap", None)

        if figcolors is not None:
            import numpy
            cmap = matplotlib.cm.get_cmap(figcolors)
            if num_colors > 1:
                colors = cmap(numpy.linspace(0, 1.0, num_colors))
            else:
                colors = [cmap(0.5)]
        else:
            from itertools import cycle
            try:

                color_cycle = cycle(matplotlib.rcParams["axes.prop_cycle"].by_key()["color"])
            except KeyError:
                color_cycle = cycle(matplotlib.rcParams["axes.color_cycle"])

            colors = [next(color_cycle) for _ in range(num_colors)]

        return colors


#===============================================================================
# INDEX: Parser for generic command-line options (e.g. warning control)
#===============================================================================


class BaseParser(Parser):
    """Parser basic options

    Parameters
    ----------
    groupname : str, optional
        Name of argument group. If not `None`, an argument group with
        the specified name will be created and added to the parser.
        If not, arguments will be in the main group.

    prefix : str, optional
        string prefix to add to default argument options (Default: "")

    disabled : list, optional
        list of parameter names that should be disabled from parser,
        without preceding dashes
    """

    def __init__(
            self,
            groupname="base_options",
            prefix="",
            disabled=None,
    ):
        """Create a parser for basic options for command-line scripts, such as
        warnings and logging

        Parameters
        ----------
        groupname : str, optional
            Name of argument group. If not `None`, an argument group with
            the specified name will be created and added to the parser.
            If not, arguments will be in the main group.

        prefix : str, optional
            string prefix to add to default argument options (Default: "")

        disabled : list, optional
            list of parameter names that should be disabled from parser,
            without preceding dashes
        """
        Parser.__init__(self, groupname=groupname, prefix=prefix, disabled=disabled)
        self.arguments = []


#         self.level_desc = ["--silent","--quiet","--verbose","--raise"]

    def get_parser(self, title=None, description=None):
        """Return an :py:class:`~argparse.ArgumentParser`

        Parameters
        ----------

        title : str, optional
            title for option group (used in command-line help screen)

        description : str, optional
            description of parser (used in command-line help screen)


        Returns
        -------
        :class:`argparse.ArgumentParser`
        """
        p = Parser.get_parser(self)
        g = p.add_argument_group(title="warning/error options")

        g.add_argument(
            "-q",
            "--quiet",
            dest="warnlevel",
            action="store_const",
            const=-1,
            help="Suppress all warning messages. Cannot use with '-v'."
        )
        g.add_argument(
            "-v",
            "--verbose",
            dest="warnlevel",
            action="count",
            help=(
                "Increase verbosity. With '-v', show every warning. "
                "With '-vv', turn warnings into exceptions. "
                "Cannot use with '-q'. "
                "(Default: show each type of warning once)"
            )
        )
        p.set_defaults(warnlevel=0)

        return p

    def get_base_ops_from_args(self, args):

        global warnings

        args = PrefixNamespaceWrapper(args, self.prefix)
        warnlevel = args.warnlevel
        actions = ["ignore", "onceperfamily", "always", "error"]

        if warnlevel >= len(actions) - 1:
            warnlevel = len(actions) - 2
        try:
            action = actions[warnlevel + 1]
        except IndexError:
            warnings.warn(
                (
                    "Invalid warning level. Expected 0-3, found %s. "
                    "Defaulting to level 1 (`--once`)." % warnlevel
                ),
                UserWarning,
            )
            action = actions[1]

        for type_, msg in PLASTID_WARNINGS:
            filterwarnings(action, message=msg, category=type_)

PLASTID_WARNINGS = [

    # mapping rules
    (DataWarning, "File contains read alignments shorter"),
    (DataWarning, "No offset for reads of length"),
    (DataWarning, "longer than read length"),

    # genome_array
    (DataWarning, "Temporarily turning off normalization"),

    # roi_tools
    (DataWarning, "is a zero-length SegmentChain. Returning 0-length count vector"),

    # metagene
    (Warning, r"IndexError finding common positions at region.*"),
    (DataWarning, "has no gene_id. Inferring gene_id"),
    (DataWarning, "has no attribute"),
    (DataWarning, "Ignoring labels"),

    # phase_by_size
    (DataWarning, "is not divisible by 3. Ignoring last partial codon."),

    # util.io.filters
    (Warning, "Could not alert listener"),

    # util.services.decorators
    (DeprecationWarning, "is deprecated and will be removed from module"),

    # gff
    (DataWarning, "because it contains exons on multiple chromosomes or strands"),
    (DataWarning, "because start or stop codons are outside exon boundaries"),
    (DataWarning, "with no `Parent` or `ID`. Ignoring."),
    (DataWarning, "because it contains exons on multiple strands"),
    (DataWarning, "because start or stop codons are outside exon boundaries."),

    # bed
    (FileFormatWarning, "Extra columns specified by."),
    (FileFormatWarning, "Are you sure this is a"),
    (FileFormatWarning, "Are you sure this BED file has extra columns"),
    (FileFormatWarning, "Maybe this BED has extra columns"),

    # gff_tokens
    (FileFormatWarning, "Found duplicate attribute key"),

    # BigBed
    (FileFormatWarning, "Could not find or could not parse autoSql declaration in BigBedFile"),

    # autoSql
    (DataWarning, "Could not convert autoSql value"),

    # psl
    (FileFormatWarning, "Rejecting line")
]

#===============================================================================
# INDEX: Deprecated alignment functions, now aliased to classes above
#===============================================================================


@deprecated(version="0.6.1", instead="AlignmentParser")
def get_alignment_file_parser(
        input_choices=("BAM", "bigwig", "bowtie", "wiggle"),
        disabled=None,
        prefix="",
        title=_DEFAULT_ALIGNMENT_FILE_PARSER_TITLE,
        description=_DEFAULT_ALIGNMENT_FILE_PARSER_DESCRIPTION,
        map_desc=_MAPPING_RULE_DESCRIPTION,
        return_subparsers=False
):
    tmp = AlignmentParser(input_choices=input_choices, prefix=prefix, disabled=disabled)
    return tmp.get_parser(title=title, description=description)


@deprecated(version="0.6.1", instead="AlignmentParser.get_genome_array_from_args()")
def get_genome_array_from_args(args, prefix="", disabled=None, printer=None):
    """Return a |GenomeArray|, |SparseGenomeArray| or |BAMGenomeArray|
    from arguments parsed by :py:func:`get_alignment_file_parser`

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        Namespace object from :py:func:`get_alignment_file_parser`

    prefix : str, optional
        string prefix to add to default argument options (Default: "")
        Must be same prefix that was added in call to :py:func:`get_alignment_file_parser`
        (Default: "")

    disabled : list, optional
        list of parameter names that were disabled when the argparser was created
        in :py:func:`get_alignment_file_parser`. (Default: ``[]``)

    printer : file-like, optional
        A stream to which stderr-like info can be written (default: |NullWriter|)


    Returns
    -------
    |GenomeArray|, |SparseGenomeArray|, or |BAMGenomeArray|


    See Also
    --------
    get_alignment_file_parser
        Function that creates :py:class:`~argparse.ArgumentParser` whose output
        :py:class:`~argparse.Namespace` is processed by this function
    """
    tmp = AlignmentParser(prefix=prefix, disabled=disabled)
    return tmp.get_genome_array_from_args(args, printer=printer)


#===============================================================================
# INDEX: deprecated annotation file parser, and helper functions
#===============================================================================


@deprecated(version="0.6.1", instead="AnnotationParser")
def get_annotation_file_parser(
        input_choices=["BED", "BigBed", "GTF2", "GFF3"],
        disabled=[],
        prefix="",
        title=_DEFAULT_ANNOTATION_PARSER_TITLE,
        description=_DEFAULT_ANNOTATION_PARSER_DESCRIPTION,
        return_subparsers=False
):
    """Return an :py:class:`~argparse.ArgumentParser` that opens
    annotation files from `BED`_, `BigBed`_, `GTF2`_, or `GFF3`_ formats

    Parameters
    ----------
    input_choices : list, optional
        list of permitted alignment file type choices.
        (Default: '["BED","BigBed","GTF2","GFF3"]'). 'PSL'_ may also be added

    disabled : list, optional
        list of parameter names that should be disabled from parser
        without preceding dashes

    prefix : str, optional
        string prefix to add to default argument options (Default: `''`)

    title : str, optional
        title for option group (used in command-line help screen)

    description : str, optional
        description of parser (used in command-line help screen)

    return_subparsers : bool, optional
        if True, additionally return a dictionary of subparser option groups,
        to which additional options may be added (Default: `False`)

    Returns
    -------
    :class:`argparse.ArgumentParser`


    See also
    --------
    get_transcripts_from_args
        function that parses the :py:class:`~argparse.Namespace` returned
        by this :py:class:`~argparse.ArgumentParser`
    """
    tmp = AnnotationParser(
        groupname="annotation_options",
        prefix=prefix,
        disabled=disabled,
        input_choices=input_choices
    )
    parser = tmp.get_parser(title, description)
    return parser


@deprecated(version="0.6.1", instead="AnnotationParser.get_transcripts_from_args()")
def get_transcripts_from_args(
        args, prefix="", disabled=[], printer=NullWriter(), return_type=None, require_sort=False
):
    """Return a list of |Transcript| objects from arguments parsed by
    :py:func:`get_annotation_file_parser`

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        Namespace object from :py:func:`get_annotation_file_parser`

    prefix : str, optional
        string prefix to add to default argument options.
        Must be same prefix that was added in call to :py:func:`get_annotation_file_parser`
        (Default: `''`)

    disabled : list, optional
        list of parameter names that were disabled when the annotation file
        parser was created by :py:func:`get_annotation_file_parser`.
        (Default: `[]`)

    printer : file-like, optional
        A stream to which stderr-like info can be written (Default: |NullWriter|)

    return_type : |SegmentChain| or subclass, optional
        Type of object to return (Default: |Transcript|)

    require_sort : bool, optional
        If True, quit if the annotation file(s) are not sorted or indexed

    Returns
    -------
    iterator
        |Transcript| objects, either in order of appearance (if input was a
        `BED`_, `BigBed`_, or `PSL`_ file), or sorted lexically by chromosome,
        start coordinate, end coordinate, and then strand (if input was `GTF2`_
        or `GFF3`_).


    See Also
    --------
    get_annotation_file_parser
        Function that creates :py:class:`argparse.ArgumentParser` whose output
        :py:class:`~argparse.Namespace` is processed by this function
    """
    tmp = AnnotationParser(groupname="annotation_options", prefix=prefix, disabled=disabled)
    return tmp.get_transcripts_from_args(
        args, printer=printer, return_type=return_type, require_sort=require_sort
    )


@deprecated(version="0.6.1", instead="AnnotationParser.get_parser()")
def get_segmentchain_file_parser(
        input_choices=["BED", "BigBed", "GTF2", "GFF3", "PSL"],
        disabled=[],
        prefix="",
        title=_DEFAULT_ANNOTATION_PARSER_TITLE,
        description=_DEFAULT_ANNOTATION_PARSER_DESCRIPTION
):
    """Create an :class:`~argparse.ArgumentParser` to open annotation files as
    |SegmentChains|

    Parameters
    ----------
    input_choices : list, optional
        list of permitted alignment file type choices
        (Default: `["BED","BigBed","GTF2","GFF3", "PSL"]`)

    disabled : list, optional
        list of parameter names that should be disabled from parser
        without preceding dashes

    prefix : str, optional
        string prefix to add to default argument options (Default: `''`)

    title : str, optional
        title for option group (used in command-line help screen)

    description : str, optional
        description of parser (used in command-line help screen)


    Returns
    -------
    :class:`argparse.ArgumentParser`


    See Also
    --------
    get_segmentchains_from_args
        function that parses the :py:class:`~argparse.Namespace` returned
        by this :py:class:`~argparse.ArgumentParser`
    """
    disabled.append([prefix + "add_three"])
    return get_annotation_file_parser(
        input_choices=input_choices,
        prefix=prefix,
        title=title,
        disabled=disabled,
        description=description
    )


@deprecated(version="0.6.1", instead="AnnotationParser.get_transcripts_from_args()")
def get_segmentchains_from_args(
        args, prefix="", disabled=[], printer=NullWriter(), require_sort=False
):
    """Return a list of |SegmentChain| objects from arguments parsed by an
    :class:`~argparse.ArgumentParser` created by :func:`get_segmentchain_file_parser`

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        Namespace object from :py:func:`get_segmentchain_file_parser`

    prefix : str, optional
        string prefix to add to default argument options.
        Must be same prefix that was added in call to
        :py:func:`get_segmentchain_file_parser` (Default: "")

    disabled : list, optional
        list of parameter names that were disabled when the annotation file
        parser was created by :py:func:`get_segmentchain_file_parser`.
        (Default: ``[]``)

    printer : file-like
        A stream to which stderr-like info can be written (Default: |NullWriter|)

    require_sort : bool, optional
        If True, quit if the annotation file(s) are not sorted or indexed


    Returns
    -------
    iterator
        sequence of |SegmentChain| objects, either in order of appearance
        (if input was a BED or PSL file), or sorted lexically by chromosome,
        start coordinate, end coordinate, and then strand (if input was) GTF or GFF


    See Also
    --------
    get_segmentchain_file_parser
        Function that creates :py:class:`argparse.ArgumentParser` whose output
        :py:class:`~argparse.Namespace` is processed by this function
    """
    disabled.append([prefix + "add_three"])
    return get_transcripts_from_args(
        args,
        prefix=prefix,
        disabled=disabled,
        printer=printer,
        return_type=SegmentChain,
        require_sort=require_sort
    )


@deprecated(version="0.6.1", instead="AnnotationParser")
def get_mask_file_parser(prefix="mask_", disabled=[]):
    """Create an :class:`~argparse.ArgumentParser` to open annotation files
    that describe regions of the genome to mask from analyses

    Parameters
    ----------
    prefix : str, optional
        Prefix to add to default argument options (Default: `'mask_'`)

    disabled : list, optional
        list of parameter names to disable from the mask file parser
        (Default: `[]`. `add_three` is always disabled.)

    Returns
    -------
    argparse.ArgumentParser

    See Also
    --------
    get_genome_hash_from_mask_args
        function that parses the :py:class:`~argparse.Namespace` returned
        by this :py:class:`~argparse.ArgumentParser`
    """
    tmp = AnnotationParser(
        groupname="%s_options" % prefix,
        prefix=prefix,
        disabled=disabled,
        input_choices=["BED", "GTF2", "GFF3", "BigBed", "PSL"]
    )
    return tmp.get_parser(_MASK_PARSER_TITLE, _MASK_PARSER_DESCRIPTION)


@deprecated(version="0.6.1", instead="AnnotationParser.get_genome_hash()")
def get_genome_hash_from_mask_args(args, prefix="mask_", printer=NullWriter()):
    """Return a |GenomeHash| of regions from command-line arguments

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        Namespace object from :py:func:`get_mask_file_parser`

    prefix : str, optional
        string prefix to add to default argument options.
        Must be same prefix that was added in call to :py:func:`get_mask_file_parser`
        (Default: "mask_")

    printer : file-like
        A stream to which stderr-like info can be written (Default: |NullWriter|)


    Returns
    -------
    |GenomeHash|
        Hashed data structure of masked genomic regions


    See Also
    --------
    get_mask_file_parser
        Function that creates :py:class:`argparse.ArgumentParser` whose output
        :py:class:`~argparse.Namespace` is processed by this function
    """
    tmp = AnnotationParser(groupname="mask_options", prefix=prefix)
    return tmp.get_genome_hash_from_args(args, printer=printer)


#===============================================================================
# INDEX: deprecated sequence file parser
#===============================================================================


@deprecated(version="0.6.1", instead="SequenceParser")
def get_sequence_file_parser(
        input_choices=("fasta", "fastq", "twobit", "genbank", "embl"),
        disabled=(),
        prefix="",
        title=_DEFAULT_SEQUENCE_PARSER_TITLE,
        description=_DEFAULT_SEQUENCE_PARSER_DESCRIPTION
):
    """Return an :py:class:`~argparse.ArgumentParser` that opens
    annotation files from `BED`_, `BigBed`_, `GTF2`_, or `GFF3`_ formats

    Parameters
    ----------
    input_choices : list, optional
        list of permitted sequence file type choices.
        (Default: '["FASTA","twobit","genbank","embl"]').

    disabled : list, optional
        list of parameter names that should be disabled from parser
        without preceding dashes

    prefix : str, optional
        string prefix to add to default argument options (Default: `''`)

    title : str, optional
        title for option group (used in command-line help screen)

    description : str, optional
        description of parser (used in command-line help screen)


    Returns
    -------
    :class:`argparse.ArgumentParser`


    See also
    --------
    get_seqdict_from_args
        function that parses the :py:class:`~argparse.Namespace` returned
        by this :py:class:`~argparse.ArgumentParser`
    """
    tmp = SequenceParser(disabled=disabled, prefix=prefix, input_choices=input_choices)
    return tmp.get_parser(title=title, description=description)


@deprecated(version="0.6.1", instead="SequenceParser.get_seqdict_from_args()")
def get_seqdict_from_args(args, index=True, prefix="", printer=NullWriter()):
    """Retrieve a dictionary-like object of sequences

    Parameters
    ----------
    args : :py:class:`argparse.Namespace`
        Namespace object from :py:func:`get_sequence_file_parser`

    prefix : str, optional
        string prefix to add to default argument options.
        Must be same prefix that was added in call to
        :py:func:`get_sequence_file_parser` (Default: "")

    index : bool, optional
        If sequence format is anything other than twobit, open with
        lazily-evaluating :func:`Bio.SeqIO.index` instead of
        :func:`Bio.SeqIO.to_dict` (Default: `True`)

    printer : file-like
        A stream to which stderr-like info can be written (Default: |NullWriter|)

    Returns
    -------
    dict-like
        Dictionary-like object mapping chromosome names to
        :class:`Bio.SeqRecord.SeqRecord`-like objects
    """
    tmp = SequenceParser(prefix=prefix)
    return tmp.get_seqdict_from_args(args, index=index, printer=printer)


#===============================================================================
# INDEX: deprecated plotting
#===============================================================================


@deprecated(version="0.6.1", instead="PlottingParser")
def get_plotting_parser(prefix="", disabled=[], title=_DEFAULT_PLOTTING_TITLE):
    """Return an :py:class:`~argparse.ArgumentParser` to control plotting

    Parameters
    ----------

    disabled : list, optional
        list of parameter names that should be disabled from parser
        without preceding dashes

    prefix : str, optional
        string prefix to add to default argument options (Default: `''`)

    title : str, optional
        title for option group (used in command-line help screen)

    description : str, optional
        description of parser (used in command-line help screen)


    Returns
    -------
    :class:`argparse.ArgumentParser`


    See also
    --------
    get_colors_from_args
        parse colors and/or colormaps from this argument parser
    """
    tmp = PlottingParser(prefix=prefix, disabled=disabled)
    return tmp.get_parser(title=title)


@deprecated(version="0.6.1", instead="PlottingParser.get_figure_from_args()")
def get_figure_from_args(args, **kwargs):
    """Return a :class:`matplotlib.figure.Figure` following arguments from
    :func:`get_plotting_parser`

    A new figure is created with parameters specified in `args`. If these are
    not found, values found in `**kwargs` will instead be used. If these are
    not found, we fall back to matplotlibrc values.

    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Namespace object from :func:`get_plotting_parser`

    kwargs : keyword arguments
        Fallback arguments for items not defined in `args`, plus any other
        keyword arguments.

    Returns
    -------
    :class:`matplotlib.figure.Figure`
        Matplotlib figure
    """
    tmp = PlottingParser()
    return tmp.get_figure_from_args(args, **kwargs)


@deprecated(version="0.6.1", instead="PlottingParser.get_colors_from_args()")
def get_colors_from_args(args, num_colors):
    """Return a list of colors from arguments parsed by a parser from
    :func:`get_plotting_parser`

    If a matplotlib colormap is specified in `args.figcolors`, colors will be
    generated from that map.

    Otherwise, if a stylesheet is specified, colors will be fetched from
    the stylesheet's color cycle.

    Otherwise, colors will be chosen from the default color cycle specified
    ``matplotlibrc``.


    Parameters
    ----------
    args : :class:`argparse.Namespace`
        Namespace object from :func:`get_plotting_parser`

    num_colors : int
        Number of colors to fetch


    Returns
    -------
    list
        List of matplotlib colors
    """
    tmp = PlottingParser()
    return tmp.get_colors_from_args(args, num_colors)


#===============================================================================
# INDEX: Utility classes
#===============================================================================


class PrefixNamespaceWrapper(object):
    """Wrapper class to facilitate processing of :py:class:`~argparse.Namespace`
    objects created by :py:func:`get_alignment_file_parser` or
    :py:func:`get_annotation_file_parser` with non-empty ``prefix`` values,
    as if no prefix had been used.

    Attributes
    ----------
    namespace : :py:class:`~argparse.Namespace`
        Result of calling :py:meth:`argparse.ArgumentParser.parse_args`

    prefix : str
        Prefix that will be prepended to names of attributes of `self.namespace`
        before they are fetched. Must match prefix that was used in creation
        of the :py:class:`argparse.ArgumentParser` that created `self.namespace`

    See Also
    --------
    get_annotation_file_parser

    get_alignment_file_parser

    get_genome_array_from_args

    get_transcripts_from_args
    """

    def __init__(self, namespace, prefix):
        """Create a |PrefixNamespaceWrapper|

        Parameters
        ----------
        namespace : :py:class:`~argparse.Namespace`
            Result of calling :py:meth:`argparse.ArgumentParser.parse_args`

        prefix : str
            Prefix that will be prepended to items from the
            :py:class:`~argparse.Namespace` before they are checked
        """
        self.namespace = namespace
        self.prefix = prefix

    def __getattr__(self, k):
        """Fetch an attribute from `self.namespace`, appending `self.prefix` to `k`
        before fetching

        Parameters
        ----------
        k : str
            Attribute to fetch
        """
        return getattr(self.namespace, "%s%s" % (self.prefix, k))


#===============================================================================
# INDEX: Utility functions
#===============================================================================


def _parse_variable_offset_file(fh):
    """Read a variable-offset text file into a dictionary.
    These text files contain two columns and are tab-delimited. The first column
    specifies the read length, or contains the special value `'default'`. The
    second column specifies the offset from the 5' end of that read length to
    use.

    Parameters
    ----------
    fh : file-like
        open filehandle pointing to data

    Returns
    -------
    dict
        dictionary mapping sequencing read lengths to their 5' offsets
    """
    my_dict = {}
    for line in fh:
        if line.startswith("length"):
            continue
        items = line.strip("\n").split("\t")
        if len(items) != 2:
            name = getattr(fh, "__name__", "Variable offset file")
            raise MalformedFileError(
                name, "More or fewer than two columns on line:\n\t%s" % line.strip("\n")
            )
        if items[0] == "length":
            continue
        key = items[0]
        try:
            key = key if key == "default" else int(key)
        except ValueError:
            name = getattr(fh, "__name__", "Variable offset file")
            raise MalformedFileError(
                name,
                "Non integer value for key '%s' on line:\n\t%s" % (key, line.strip("\n")),
            )
        if key in my_dict:
            name = getattr(fh, "__name__", "Variable offset file")
            raise MalformedFileError(name, "multiple offsets defined for read length %s" % key)
        else:
            try:
                my_dict[key] = int(items[1])
            except ValueError:
                name = getattr(fh, "__name__", "Variable offset file")
                raise MalformedFileError(
                    name,
                    (
                        "Non integer value for value '%s' on line:\n\t%s" %
                        (items[1], line.strip("\n"))
                    ),
                )

    return my_dict



#!/usr/bin/env python
"""Miscellaneous functions that don't seem to fit elsewhere

:py:func:`guess_formatter`
    Convert string into primitive data type it probably encoded before `str`
    was called on it

:py:func:`number`
    Convert string into numerical data type, trying `bool`, `int`, `float`,
    then `str`
"""
import numpy


def guess_formatter(inp):
    """Guesses the format of input, trying `bool`, `int`, `float`, then `str`.
    Correctly parses `nan`s and `Inf`s. Converts `None` to `nan`


    Parameters
    ----------
    inp : str
        input


    Returns
    -------
    boolean, number, or string
    """
    if inp.lower() == "true":
        return True
    elif inp.lower() == "false":
        return False
    else:
        try:
            return number(inp)
        except ValueError:
            return str(inp)


def number(inp):
    """Parses numbers from strings, preferring int over float.
    Parses `nan`, `Nan`, `None`, `none`, `inf`, and `-inf`


    Parameters
    ----------
    inp : str
        string input


    Returns
    -------
    float, numpy.nan, numpy.inf, or -numpy.inf, or str if no conversion found


    Raises
    ------
    ValueError
        if `inp` cannot be converted to a number
    """
    if inp in ("nan", "NaN", "na", "None", "none"):
        return numpy.nan
    elif inp in ("inf", "Inf"):
        return numpy.inf
    elif inp in ("-inf", "-Inf"):
        return -numpy.inf
    else:
        try:
            # note: in python bools are also ints!
            # isinstance(True,int) == True
            val = int(inp)
        except ValueError:
            val = float(inp)
        return val



#!/usr/bin/env python
"""This is a template for a command-line script. To write a script, edit this file,
making changes as necessary, Make changes below as necessary, and, when finished,
update this docstring.
"""
import inspect
import sys
import argparse
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.io.filters import NameDateWriter
from plastid.util.io.openers import get_short_name

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: sys.argv[1:] (actual command-line arguments)
    """
    # fill out command-line program here

    # add parents from plastid.scriptlib.argparsers as necessary
    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[]
    )

    # add your own argument
    parser.add_argument("--foo", type=str, help="Some argument foo")

    args = parser.parse_args(argv)

    # write program body
    pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Welcome to plastid!

This package contains various utilities for analyzing high-throughput sequencing
data, with an emphasis on simplicity for users. To this end, this package provides:

  #. A set of command-line scripts that implement common sequencing workflows
     (see |bin|).
  
  #. Readers that abstract data from various file formats into a minimal set of
     object types. These object types define APIs that easily interface with
     existing scientific tools, such as the `SciPy`_ stack (see |genomics| and
     |readers|)

  #. Tools to facilitate writing command-line scripts (see |scriptlib|)


Package overview
----------------
plastid is divided into the following subpackages:

    ==============    =========================================================
    Package           Contents
    --------------    ---------------------------------------------------------
    |bin|             Command-line scripts
    |genomics|        Classes and functions to manipulate genome annotations, alignments, and quantitative data
    |plotting|        Tools for plotting
    |readers|         Parsers for various file formats
    |util|            Utilities (e.g. function decorators, exceptions, argument parsers)
    |test|            Unit and functional tests (requires download of test datasets)
    ==============    =========================================================
     
"""
__version__ = "0.6.1"
__author__ = "Joshua Griffin Dunn"
import matplotlib
matplotlib.use("agg")

from plastid.genomics.roitools import (
    GenomicSegment,
    SegmentChain,
    Transcript,
)

from plastid.genomics.genome_array import (
    BAMGenomeArray,
    BigWigGenomeArray,
    GenomeArray,
    SparseGenomeArray,
    variable_five_prime_map,
    five_prime_map,
    center_map,
    three_prime_map,
)

from plastid.genomics.genome_hash import (
    GenomeHash,
    TabixGenomeHash,
    BigBedGenomeHash,
)

from plastid.genomics.map_factories import (
    StratifiedVariableFivePrimeMapFactory,
    VariableFivePrimeMapFactory,
    FivePrimeMapFactory,
    CenterMapFactory,
    ThreePrimeMapFactory,
    SizeFilterFactory,
)

from plastid.readers.bed import BED_Reader
from plastid.readers.bigbed import BigBedReader
from plastid.readers.gff import (
    GTF2_Reader,
    GFF3_Reader,
    GTF2_TranscriptAssembler,
    GFF3_TranscriptAssembler,
)

from plastid.readers.psl import PSL_Reader
from plastid.readers.bigwig import BigWigReader
from plastid.util.io.openers import read_pl_table, write_pl_table
from plastid.util.services.exceptions import formatwarning
